from domain.models.base_model import ExternalUrl, QueryParams
from domain.constants.institutions import institutions_list
from domain.parsers import affiliation_parser
from domain.models.affiliation_model import Affiliation, Relation
from infrastructure.repositories import (
    person_repository,
    affiliation_repository,
)


def get_affiliation_by_id(affiliation_id: str, affiliation_type: str) -> dict:
    affiliation = affiliation_repository.get_affiliation_by_id(affiliation_id)
    set_relation_external_urls(affiliation)
    set_upper_affiliations_and_logo(affiliation, affiliation_type)
    return {"data": affiliation.model_dump(by_alias=True)}


def get_related_affiliations_by_affiliation(affiliation_id: str, affiliation_type: str) -> dict:
    data = {}
    if affiliation_type == "institution":
        faculties = affiliation_repository.get_affiliations_by_institution(affiliation_id, "faculty")
        departments = affiliation_repository.get_affiliations_by_institution(affiliation_id, "department")
        groups = affiliation_repository.get_affiliations_by_institution(affiliation_id, "group")
        data["faculties"] = [faculty.model_dump(include={"id", "name"}) for faculty in faculties]
        data["departments"] = [department.model_dump(include={"id", "name"}) for department in departments]
        data["groups"] = [group.model_dump(include={"id", "name"}) for group in groups]
    elif affiliation_type == "faculty":
        departments = affiliation_repository.get_departments_by_faculty(affiliation_id)
        groups = affiliation_repository.get_groups_by_faculty_or_department(affiliation_id)
        authors = person_repository.get_persons_by_affiliation(affiliation_id)
        data["departments"] = [department.model_dump(include={"id", "name"}) for department in departments]
        data["groups"] = [group.model_dump(include={"id", "name"}) for group in groups]
        data["authors"] = [author.model_dump(include={"id", "full_name"}) for author in authors]
    elif affiliation_type == "department":
        groups = affiliation_repository.get_groups_by_faculty_or_department(affiliation_id)
        authors = person_repository.get_persons_by_affiliation(affiliation_id)
        data["groups"] = [group.model_dump(include={"id", "name"}) for group in groups]
        data["authors"] = [author.model_dump(include={"id", "full_name"}) for author in authors]
    elif affiliation_type == "group":
        authors = person_repository.get_persons_by_affiliation(affiliation_id)
        data["authors"] = [author.model_dump(include={"id", "full_name"}) for author in authors]
    return data


def search_affiliations(affiliation_type: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "names",
            "addresses.country_code",
            "external_ids",
            "external_urls",
            "relations",
            "types",
            "citations_count",
            "products_count",
            "relations_data",
        ]
    }
    affiliations, total_results = affiliation_repository.search_affiliations(
        affiliation_type, query_params, pipeline_params
    )
    affiliations_list = []
    for affiliation in affiliations:
        set_relation_external_urls(affiliation)
        set_upper_affiliations_and_logo(affiliation, affiliation_type)
        affiliations_list.append(affiliation)
    data = affiliation_parser.parse_search_result(affiliations_list)
    return {"data": data, "total_results": total_results}


def set_relation_external_urls(affiliation: Affiliation) -> None:
    for relation in affiliation.relations:
        relation_data = next(
            filter(lambda x: x.id == relation.id, affiliation.relations_data),
            Relation(),
        )
        relation.external_urls = relation_data.external_urls


def set_upper_affiliations_and_logo(affiliation: Affiliation, affiliation_type: str) -> None:
    if affiliation_type == "institution":
        affiliation.logo = next(
            filter(lambda x: x.source == "logo", affiliation.external_urls),
            ExternalUrl(),
        ).url
    upper_affiliations = []
    for relation in affiliation.relations:
        if affiliation_type == "faculty" and relation.types[0].type in institutions_list:
            set_logo(affiliation, relation)
            upper_affiliations.append(relation)
        elif affiliation_type == "department" and relation.types[0].type in institutions_list + ["faculty"]:
            set_logo(affiliation, relation)
            upper_affiliations.append(relation)
        elif affiliation_type == "group" and relation.types[0].type in institutions_list + [
            "department",
            "faculty",
        ]:
            set_logo(affiliation, relation)
            upper_affiliations.append(relation)
    affiliation.affiliations = upper_affiliations


def set_logo(affiliation: Affiliation, relation: Relation) -> None:
    if relation.types[0].type in institutions_list:
        affiliation.logo = next(filter(lambda x: x.source == "logo", relation.external_urls), ExternalUrl()).url



from datetime import datetime
from typing import Generator

from domain.models.base_model import ExternalId, QueryParams
from domain.models.work_model import Work, Affiliation
from infrastructure.repositories import csv_repository
from domain.constants.institutions import institutions_list
from domain.constants.openalex_types import openalex_types_dict
from domain.services import source_service
from domain.services.work_service import set_title_and_language
from domain.parsers import work_parser


def get_works_csv_by_affiliation(affiliation_id: str, query_params: QueryParams) -> str:
    works = csv_repository.get_works_csv_by_affiliation(affiliation_id, query_params)
    data = get_csv_data(works)
    return work_parser.parse_csv(data)


def get_works_csv_by_person(person_id: str, query_params: QueryParams) -> str:
    works = csv_repository.get_works_csv_by_person(person_id, query_params)
    data = get_csv_data(works)
    return work_parser.parse_csv(data)


def get_csv_data(works: Generator) -> list:
    data = []
    for work in works:
        set_open_access_status(work)
        set_doi(work)
        set_csv_ranking(work)
        set_csv_affiliations(work)
        set_csv_authors(work)
        set_csv_bibliographic_info(work)
        set_csv_citations_count(work)
        set_csv_subjects(work)
        set_title_and_language(work)
        set_csv_types(work)
        source_service.update_csv_work_source(work)
        data.append(work)
    return data


def set_open_access_status(work: Work) -> None:
    if work.open_access:
        work.open_access_status = work.open_access.open_access_status


def set_csv_ranking(work: Work) -> None:
    if work.ranking:
        rankings = []
        for ranking in work.ranking:
            if type(ranking.date) == int:
                date = datetime.fromtimestamp(ranking.date).strftime("%d-%m-%Y")
                rankings.append(str(ranking.rank) + " / " + str(ranking.source) + " / " + str(date))
            else:
                rankings.append(str(ranking.rank) + " / " + str(ranking.source))
        work.ranking = " | ".join(set(rankings))
    else:
        work.ranking = None


def set_doi(work: Work) -> None:
    work.doi = next(
        filter(lambda external_id: external_id.source == "doi", work.external_ids),
        ExternalId(),
    ).id


def set_csv_types(work: Work) -> None:
    openalex_types = []
    scienti_types = []
    for work_type in work.types:
        if work_type.source == "openalex" and work_type.type in openalex_types_dict.keys():
            openalex_types.append(openalex_types_dict.get(work_type.type))
        elif work_type.source == "scienti":
            scienti_types.append(str(work_type.type))
    work.openalex_types = " | ".join(set(openalex_types))
    work.scienti_types = " | ".join(set(scienti_types))


def set_csv_subjects(work: Work) -> None:
    if work.subjects:
        subjects = []
        for subject in work.subjects[0].subjects:
            subjects.append(str(subject.name))
        work.subjects = " | ".join(set(subjects))


def set_csv_citations_count(work: Work) -> None:
    for citation_count in work.citations_count:
        if citation_count.source == "openalex":
            work.openalex_citations_count = str(citation_count.count)
        elif citation_count.source == "scholar":
            work.scholar_citations_count = str(citation_count.count)


def set_csv_bibliographic_info(work: Work) -> None:
    work.bibtex = work.bibliographic_info.bibtex
    work.pages = work.bibliographic_info.pages
    work.issue = work.bibliographic_info.issue
    work.start_page = work.bibliographic_info.start_page
    work.end_page = work.bibliographic_info.end_page
    work.volume = work.bibliographic_info.volume


def set_csv_authors(work: Work) -> None:
    authors = []
    for author in work.authors:
        authors.append(str(author.full_name))
    work.authors = " | ".join(set(authors))


def set_csv_affiliations(work: Work) -> None:
    countries = []
    institutions = []
    departments = []
    faculties = []
    groups = []
    groups_ranking = []
    for author in work.authors:
        for affiliation in author.affiliations:
            affiliation_data = next(
                filter(lambda x: x.id == affiliation.id, work.affiliations_data),
                Affiliation(),
            )
            if affiliation.types and affiliation.types[0].type in institutions_list:
                institutions.append(str(affiliation.name))
                if affiliation_data.addresses:
                    countries.append(str(affiliation_data.addresses[0].country))
            elif affiliation.types and affiliation.types[0].type == "department":
                departments.append(str(affiliation.name))
            elif affiliation.types and affiliation.types[0].type == "faculty":
                faculties.append(str(affiliation.name))
            elif affiliation.types and affiliation.types[0].type == "group":
                groups.append(str(affiliation.name))
                if affiliation_data.ranking:
                    ranking = affiliation_data.ranking[0]
                    if type(ranking.from_date) == int and type(ranking.to_date) == int:
                        groups_ranking.append(
                            str(ranking.rank)
                            + " / "
                            + datetime.fromtimestamp(ranking.from_date).strftime("%d-%m-%Y")
                            + " - "
                            + datetime.fromtimestamp(ranking.to_date).strftime("%d-%m-%Y")
                        )
    work.institutions = " | ".join(set(institutions))
    work.departments = " | ".join(set(departments))
    work.faculties = " | ".join(set(faculties))
    work.groups = " | ".join(set(groups))
    work.groups_ranking = " | ".join(set(groups_ranking))
    work.countries = " | ".join(set(countries))



from infrastructure.repositories import (
    info_repository,
)


def get_info() -> dict:
    return {
        "db_update": info_repository.get_last_db_update(),
        "total_products": info_repository.get_entity_count("works"),
        "total_institutions": info_repository.get_entity_count("affiliations", "institution"),
        "total_faculties": info_repository.get_entity_count("affiliations", "faculty"),
        "total_departments": info_repository.get_entity_count("affiliations", "department"),
        "total_groups": info_repository.get_entity_count("affiliations", "group"),
        "total_patents": info_repository.get_entity_count("patents"),
        "total_projects": info_repository.get_entity_count("projects"),
        "total_other_products": info_repository.get_entity_count("works_misc"),
        "total_authors": info_repository.get_entity_count("person"),
    }



from pymongo.command_cursor import CommandCursor

from domain.constants.articles_types import articles_types_list
from domain.models.base_model import QueryParams
from infrastructure.repositories import (
    work_repository,
    plot_repository,
    calculations_repository,
    affiliation_repository,
)
from domain.parsers import pie_parser, map_parser, venn_parser, bar_parser, network_parser


def get_affiliation_plot(affiliation_id: str, affiliation_type: str, query_params: QueryParams) -> dict | None:
    plot_type = query_params.plot
    plot_type_dict = {
        "faculties_by_product_type": "faculty",
        "departments_by_product_type": "department",
        "research_groups_by_product_type": "group",
    }
    if plot_type in plot_type_dict.keys():
        relation_type = plot_type_dict[plot_type]
        return plot_affiliations_by_product_type(affiliation_id, affiliation_type, relation_type, query_params)
    if plot_type in [
        "citations_by_faculty",
        "citations_by_department",
        "citations_by_research_group",
    ]:
        relation_type = plot_type.split("_")[-1]
        return plot_citations_by_affiliations(affiliation_id, affiliation_type, relation_type)
    if plot_type in [
        "apc_expenses_by_faculty",
        "apc_expenses_by_department",
        "apc_expenses_by_group",
    ]:
        relation_type = plot_type.split("_")[-1]
        return plot_apc_expenses_by_affiliation(affiliation_id, affiliation_type, relation_type, query_params)
    if plot_type in [
        "h_index_by_faculty",
        "h_index_by_department",
        "h_index_by_research_group",
    ]:
        relation_type = plot_type.split("_")[-1]
        return plot_h_index_by_affiliation(affiliation_id, affiliation_type, relation_type, query_params)
    return globals()["plot_" + plot_type](affiliation_id, query_params)


def plot_affiliations_by_product_type(
    affiliation_id: str, affiliation_type: str, relation_type: str, query_params: QueryParams
) -> dict | None:
    data: CommandCursor | None = None
    if affiliation_type == "institution":
        data = plot_repository.get_affiliations_scienti_works_count_by_institution(
            affiliation_id, relation_type, query_params
        )
    elif affiliation_type == "faculty" and relation_type == "department":
        data = plot_repository.get_departments_scienti_works_count_by_faculty(affiliation_id, query_params)
    elif affiliation_type in ["faculty", "department"] and relation_type == "group":
        data = plot_repository.get_groups_scienti_works_count_by_faculty_or_department(affiliation_id, query_params)
    return bar_parser.parse_affiliations_by_product_type(data)


def plot_citations_by_affiliations(affiliation_id: str, affiliation_type: str, relation_type: str) -> dict:
    data: CommandCursor | None = None
    if affiliation_type == "institution":
        data = plot_repository.get_affiliations_citations_count_by_institution(affiliation_id, relation_type)
    elif affiliation_type == "faculty" and relation_type == "department":
        data = plot_repository.get_departments_citations_count_by_faculty(affiliation_id)
    elif affiliation_type in ["faculty", "department"] and relation_type == "group":
        data = plot_repository.get_groups_citations_count_by_faculty_or_department(affiliation_id)
    return pie_parser.parse_citations_by_affiliations(data)


def plot_apc_expenses_by_affiliation(
    affiliation_id: str, affiliation_type: str, relation_type: str, query_params: QueryParams
) -> dict:
    data: CommandCursor | None = None
    if affiliation_type == "institution":
        data = plot_repository.get_affiliations_apc_expenses_by_institution(affiliation_id, relation_type, query_params)
    elif affiliation_type == "faculty" and relation_type == "department":
        data = plot_repository.get_departments_apc_expenses_by_faculty(affiliation_id, query_params)
    elif affiliation_type in ["faculty", "department"] and relation_type == "group":
        data = plot_repository.get_groups_apc_expenses_by_faculty_or_department(affiliation_id, query_params)
    return pie_parser.parse_apc_expenses_by_affiliations(data)


def plot_h_index_by_affiliation(
    affiliation_id: str, affiliation_type: str, relation_type: str, query_params: QueryParams
) -> dict:
    data: CommandCursor | None = None
    if affiliation_type == "institution":
        data = plot_repository.get_affiliations_works_citations_count_by_institution(
            affiliation_id, relation_type, query_params
        )
    elif affiliation_type == "faculty" and relation_type == "department":
        data = plot_repository.get_departments_works_citations_count_by_faculty(affiliation_id, query_params)
    elif affiliation_type in ["faculty", "department"] and relation_type == "group":
        data = plot_repository.get_groups_works_citations_count_by_faculty_or_department(affiliation_id, query_params)
    return pie_parser.parse_h_index_by_affiliation(data)


def plot_annual_evolution_by_scienti_classification(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"project": ["year_published", "types"]}
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return bar_parser.parse_annual_evolution_by_scienti_classification(works)


def plot_annual_citation_count(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"project": ["citations_by_year"]}
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return bar_parser.parse_annual_citation_count(works)


def plot_annual_articles_open_access(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "project": ["year_published", "open_access"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return bar_parser.parse_annual_articles_open_access(works)


def plot_annual_articles_by_top_publishers(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["publisher"],
        "work_project": ["source", "year_published", "types"],
        "match": {
            "types.type": {"$in": articles_types_list},
            "source.publisher.name": {"$ne": float("nan")},
        },
    }
    works = work_repository.get_works_with_source_by_affiliation(affiliation_id, query_params, pipeline_params)
    return bar_parser.parse_annual_articles_by_top_publishers(works)


def plot_most_used_title_words(affiliation_id: str, query_params: QueryParams) -> dict:
    data = calculations_repository.get_affiliation_calculations(affiliation_id)
    return pie_parser.parse_most_used_title_words(data)


def plot_articles_by_publisher(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["publisher"],
        "work_project": ["source"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_publisher(works)


def plot_products_by_subject(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"subjects": {"$ne": []}},
        "project": ["subjects"],
    }
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_products_by_subject(works)


def plot_products_by_database(affiliation_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_products_by_database_by_affiliation(affiliation_id)
    return venn_parser.parse_products_by_database(data)


def plot_articles_by_access_route(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"types.type": {"$in": articles_types_list}},
        "project": ["open_access"],
    }
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_products_by_access_route(works)


def plot_active_authors_by_sex(affiliation_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_active_authors_by_sex(affiliation_id)
    return pie_parser.parse_active_authors_by_sex(data)


def plot_active_authors_by_age_range(affiliation_id: str, query_params: QueryParams) -> dict:
    persons = plot_repository.get_active_authors_by_age_range(affiliation_id)
    return pie_parser.parse_active_authors_by_age_range(persons)


def plot_articles_by_scienti_category(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"types.type": {"$in": articles_types_list}},
        "project": ["ranking"],
    }
    works = work_repository.get_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_scienti_category(list(works))


def plot_articles_by_scimago_quartile(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["ranking"],
        "work_project": ["source", "date_published"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_scimago_quartile(works)


def plot_articles_by_publishing_institution(affiliation_id: str, query_params: QueryParams) -> dict:
    institution = affiliation_repository.get_affiliation_by_id(affiliation_id)
    pipeline_params = {
        "source_project": ["publisher"],
        "work_project": ["source"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_affiliation(affiliation_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_publishing_institution(works, institution)


def plot_coauthorship_by_country_map(affiliation_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_coauthorship_by_country_map_by_affiliation(affiliation_id, query_params)
    return map_parser.parse_coauthorship_by_country_map(data)


def plot_coauthorship_by_colombian_department_map(affiliation_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_coauthorship_by_colombian_department_map_by_affiliation(affiliation_id, query_params)
    return map_parser.get_coauthorship_by_colombian_department_map(data)


def plot_institutional_coauthorship_network(affiliation_id: str, query_params: QueryParams) -> dict:
    data = calculations_repository.get_affiliation_calculations(affiliation_id)
    return network_parser.parse_institutional_coauthorship_network(data)


def plot_annual_apc_expenses(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"source_project": ["apc"], "work_project": ["source", "year_published"]}
    works = work_repository.get_works_with_source_by_affiliation(affiliation_id, query_params, pipeline_params)
    return bar_parser.parse_annual_apc_expenses(works)



from typing import Tuple

from flask import Blueprint, jsonify, Response
from sentry_sdk import capture_exception

from domain.services import work_service

work_app_router = Blueprint("work_app_router", __name__)

"""
@api {get} /app/works/:work_id Get work by id
@apiName GetWorkById
@apiGroup Work
@apiVersion 1.0.0
@apiDescription Obtiene un producto bibliográfico por ID.

@apiParam {String} work_id ID del producto bibliográfico.
"""


@work_app_router.route("/<work_id>", methods=["GET"])
def get_work_by_id(work_id: str) -> Response | Tuple[Response, int]:
    data = work_service.get_work_by_id(work_id)
    return jsonify(data)


"""
@api {get} /app/works/:work_id/authors Get work authors
@apiName GetWorkAuthors
@apiGroup Work
@apiVersion 1.0.0
@apiDescription Obtiene los autores de un producto bibliográfico.

@apiParam {String} work_id ID del producto bibliográfico.
"""


@work_app_router.route("/<work_id>/authors", methods=["GET"])
def get_work_authors(work_id: str) -> Response | Tuple[Response, int]:
    try:
        data = work_service.get_work_authors(work_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from typing import Tuple

from flask import Blueprint, jsonify, Response
from sentry_sdk import capture_exception

from domain.services import patent_service

patent_app_router = Blueprint("patent_app_router", __name__)

"""
@api {get} /app/patent/:patent_id Get patent by id
@apiName GetPatentById
@apiGroup Patent
@apiVersion 1.0.0
@apiDescription Obtiene una patente por ID.

@apiParam {String} patent_id ID de la patente.
"""


@patent_app_router.route("/<patent_id>", methods=["GET"])
def get_patent_by_id(patent_id: str) -> Response | Tuple[Response, int]:
    try:
        data = patent_service.get_patent_by_id(patent_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/patent/:patent_id/authors Get patent authors
@apiName GetPatentAuthors
@apiGroup Patent
@apiVersion 1.0.0
@apiDescription Obtiene los autores de una patente.

@apiParam {String} patent_id ID del patente.
"""


@patent_app_router.route("/<patent_id>/authors", methods=["GET"])
def get_patent_authors(patent_id: str) -> Response | Tuple[Response, int]:
    try:
        data = patent_service.get_patent_authors(patent_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from typing import Tuple

from flask import Blueprint, jsonify, Response
from sentry_sdk import capture_exception

from domain.services import other_work_service

other_work_app_router = Blueprint("other_work_app_router", __name__)

"""
@api {get} /app/other_work/:other_work_id Get other work by id
@apiName GetOtherWorkById
@apiGroup Other Work
@apiVersion 1.0.0
@apiDescription Obtiene un producto de otro tipo por ID.

@apiParam {String} other_work_id ID del producto de otro tipo.
"""


@other_work_app_router.route("/<other_work_id>", methods=["GET"])
def get_other_work_by_id(other_work_id: str) -> Response | Tuple[Response, int]:
    try:
        data = other_work_service.get_other_work_by_id(other_work_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/other_work/:other_work_id/authors Get other work authors
@apiName GetOtherWorkAuthors
@apiGroup Other Work
@apiVersion 1.0.0
@apiDescription Obtiene los autores de un producto de otro tipo.

@apiParam {String} other_work_id ID del producto de otro tipo.
"""


@other_work_app_router.route("/<other_work_id>/authors", methods=["GET"])
def get_other_work_authors(other_work_id: str) -> Response | Tuple[Response, int]:
    try:
        data = other_work_service.get_other_work_authors(other_work_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from bson import ObjectId
from pydantic import BaseModel, Field

from domain.models.base_model import (
    PyObjectId,
    Type,
    ExternalId,
    ExternalUrl,
    Updated,
    Ranking,
    Author,
    Group,
    Title,
    ProductType,
)


class Project(BaseModel):
    id: PyObjectId = Field(alias="_id")
    abstract: str | None = None
    authors_count: int | None = Field(default_factory=int, alias="author_count")
    authors: list[Author] | str | None = None
    date_init: int | None = None
    date_end: int | None = None
    external_ids: list[ExternalId] | None = None
    external_urls: list[ExternalUrl] | None = None
    groups: list[Group] | str | None = None
    keywords: list | None = None
    ranking: list[Ranking] | str | None = None
    titles: list[Title] | None = None
    types: list[Type] | None = None
    updated: list[Updated] | None = None
    year_init: int | str | None = None
    year_end: int | str | None = None

    authors_data: list[Author] | None = None
    language: str | None = None
    product_types: list[ProductType] | None = None
    title: str | None = None

    class Config:
        json_encoders = {ObjectId: str}



from bson import ObjectId

from domain.models.source_model import Source
from infrastructure.mongo import database
from domain.exceptions.not_entity_exception import NotEntityException


def get_source_by_id(source_id: str) -> Source:
    source_data = database["sources"].find_one({"_id": ObjectId(source_id)})
    if not source_data:
        raise NotEntityException(f"The source with id {source_id} does not exist.")
    return Source(**source_data)



from flask import Blueprint

from application.routes.api.apc_api_router import apc_api_router
from application.routes.app.info_app_router import info_app_router
from config import settings
from application.routes.app.other_work_app_router import other_work_app_router
from application.routes.app.patent_app_router import patent_app_router
from application.routes.app.project_app_router import project_app_router
from application.routes.app.search_app_router import search_app_router
from application.routes.api.search_api_router import search_api_router
from application.routes.app.affiliation_app_router import affiliation_app_router
from application.routes.api.affiliation_api_router import affiliation_api_router
from application.routes.app.person_app_router import person_app_router
from application.routes.api.person_api_router import person_api_router
from application.routes.app.work_app_router import work_app_router
from application.routes.docs_router import router as docs_router
from application.routes.ping_router import ping_router

router = Blueprint("router", __name__)

router.register_blueprint(ping_router)
router.register_blueprint(docs_router)

router.register_blueprint(info_app_router, url_prefix=f"{settings.APP_URL_PREFIX}")

router.register_blueprint(search_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/search")
router.register_blueprint(search_api_router, url_prefix=f"{settings.API_URL_PREFIX}/search")

router.register_blueprint(affiliation_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/affiliation")
router.register_blueprint(affiliation_api_router, url_prefix=f"{settings.API_URL_PREFIX}/affiliation")

router.register_blueprint(person_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/person")
router.register_blueprint(person_api_router, url_prefix=f"{settings.API_URL_PREFIX}/person")

router.register_blueprint(work_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/work")

router.register_blueprint(other_work_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/other_work")

router.register_blueprint(patent_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/patent")

router.register_blueprint(project_app_router, url_prefix=f"{settings.APP_URL_PREFIX}/project")

router.register_blueprint(apc_api_router, url_prefix=f"{settings.API_URL_PREFIX}/apc")



from typing import Generator

from domain.models.base_model import QueryParams
from infrastructure.repositories import other_work_repository
from domain.services.base_service import (
    set_external_ids,
    set_external_urls,
    set_authors_external_ids,
    limit_authors,
    set_product_types,
    set_title_and_language,
)
from domain.parsers import other_work_parser


def get_other_work_by_id(other_work_id: str) -> dict:
    other_work = other_work_repository.get_other_work_by_id(other_work_id)
    set_external_ids(other_work)
    set_external_urls(other_work)
    limit_authors(other_work)
    set_authors_external_ids(other_work)
    set_title_and_language(other_work)
    set_product_types(other_work)
    data = other_work_parser.parse_other_work(other_work)
    return {"data": data}


def get_other_work_authors(other_work_id: str) -> dict:
    other_work = other_work_repository.get_other_work_by_id(other_work_id)
    set_authors_external_ids(other_work)
    return {"data": other_work.model_dump()["authors"]}


def search_other_works(query_params: QueryParams) -> dict:
    pipeline_params = get_other_works_by_entity_pipeline_params()
    other_works, total_results = other_work_repository.search_other_works(query_params, pipeline_params)
    other_works_data = get_other_work_by_entity_data(other_works)
    data = other_work_parser.parse_search_results(other_works_data)
    return {"data": data, "total_results": total_results}


def get_other_works_by_affiliation(affiliation_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_other_works_by_entity_pipeline_params()
    other_works = other_work_repository.get_other_works_by_affiliation(affiliation_id, query_params, pipeline_params)
    other_works_data = get_other_work_by_entity_data(other_works)
    data = other_work_parser.parse_other_works_by_entity(other_works_data)
    total_results = other_work_repository.get_other_works_count_by_affiliation(affiliation_id)
    return {"data": data, "total_results": total_results}


def get_other_works_by_person(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = get_other_works_by_entity_pipeline_params()
    other_works = other_work_repository.get_other_works_by_person(person_id, query_params, pipeline_params)
    other_works_data = get_other_work_by_entity_data(other_works)
    data = other_work_parser.parse_other_works_by_entity(other_works_data)
    total_results = other_work_repository.get_other_works_count_by_person(person_id)
    return {"data": data, "total_results": total_results}


def get_other_work_by_entity_data(other_works: Generator) -> list:
    other_works_data = []
    for other_work in other_works:
        limit_authors(other_work)
        set_title_and_language(other_work)
        set_product_types(other_work)
        other_works_data.append(other_work)
    return other_works_data


def get_other_works_by_entity_pipeline_params() -> dict:
    pipeline_params = {
        "project": [
            "_id",
            "author_count",
            "authors",
            "types",
            "titles",
            "subjects",
            "year_published",
            "external_ids",
            "external_urls",
            "authors_data",
        ]
    }
    return pipeline_params



from typing import Generator

from pymongo.command_cursor import CommandCursor

from domain.models.other_work_model import OtherWork


def get(cursor: CommandCursor) -> Generator:
    for document in cursor:
        yield OtherWork(**document)



from typing import Generator, Tuple

from bson import ObjectId
from pymongo.command_cursor import CommandCursor

from domain.models.base_model import QueryParams
from infrastructure.generators import work_generator
from infrastructure.mongo import database, calculations_database
from infrastructure.repositories import work_repository


def get_affiliations_scienti_works_count_by_institution(
    institution_id: str, relation_type: str, query_params: QueryParams
) -> CommandCursor:
    pipeline = [
        {"$match": {"relations.id": institution_id, "types.type": relation_type}},
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [{"$project": {"types": 1}}],
            }
        },
        {"$unwind": "$works"},
        {"$unwind": "$works.types"},
        {"$match": {"works.types.source": "scienti", "works.types.level": 2}},
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$group": {
                "_id": {"id": "$_id", "type": "$works.types.type", "name": "$names.name"},
                "works_count": {"$sum": 1},
            }
        },
        {"$project": {"_id": 0, "type": "$_id.type", "works_count": 1, "name": {"$first": "$_id.name"}}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_departments_scienti_works_count_by_faculty(affiliation_id: str, query_params: QueryParams) -> CommandCursor:
    return get_affiliations_scienti_works_count_by_institution(affiliation_id, "department", query_params)


def get_groups_scienti_works_count_by_faculty_or_department(
    affiliation_id: str, query_params: QueryParams
) -> CommandCursor:
    institution_id = (
        database["affiliations"]
        .aggregate(
            [
                {"$match": {"_id": affiliation_id}},
                {"$unwind": "$relations"},
                {"$match": {"relations.types.type": "education"}},
            ]
        )
        .next()
        .get("relations", {})
        .get("id", None)
    )
    pipeline = [
        {
            "$match": {
                "relations.id": institution_id,
                "types.type": "group",
            }
        },
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [{"$project": {"types": 1}}],
            }
        },
        {"$unwind": "$works"},
        {"$unwind": "$works.types"},
        {
            "$match": {
                "works.authors.affiliations.id": affiliation_id,
                "works.types.source": "scienti",
                "works.types.level": 2,
            }
        },
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$group": {
                "_id": {"id": "$_id", "type": "$works.types.type", "name": "$names.name"},
                "works_count": {"$sum": 1},
            }
        },
        {"$project": {"_id": 0, "type": "$_id.type", "works_count": 1, "name": {"$first": "$_id.name"}}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_affiliations_citations_count_by_institution(institution_id: str, relation_type: str) -> CommandCursor:
    pipeline = [
        {"$match": {"relations.id": institution_id, "types.type": relation_type}},
        {"$project": {"_id": 0, "citations_count": 1, "name": {"$first": "$names.name"}}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_departments_citations_count_by_faculty(affiliation_id: str) -> CommandCursor:
    return get_affiliations_citations_count_by_institution(affiliation_id, "department")


def get_groups_citations_count_by_faculty_or_department(affiliation_id: str) -> CommandCursor:
    institution_id = (
        database["affiliations"]
        .aggregate(
            [
                {"$match": {"_id": affiliation_id}},
                {"$unwind": "$relations"},
                {"$match": {"relations.types.type": "education"}},
            ]
        )
        .next()
        .get("relations", {})
        .get("id", None)
    )
    pipeline = [
        {
            "$match": {
                "authors.affiliations.id": affiliation_id,
            }
        },
        {"$unwind": "$groups"},
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "groups.id",
                "foreignField": "_id",
                "as": "group",
                "pipeline": [
                    {
                        "$match": {
                            "relations.id": institution_id,
                            "types.type": "group",
                        },
                    },
                    {"$project": {"_id": 1, "name": {"$first": "$names.name"}, "citations_count": 1}},
                ],
            }
        },
        {"$unwind": "$group"},
        {"$replaceRoot": {"newRoot": "$group"}},
        {
            "$group": {
                "_id": "$_id",
                "name": {"$first": "$name"},
                "citations_count": {"$first": "$citations_count"},
            }
        },
    ]
    return database["works"].aggregate(pipeline)


def get_affiliations_apc_expenses_by_institution(
    institution_id: str, relation_type: str, query_params: QueryParams
) -> CommandCursor:
    pipeline = [
        {"$match": {"relations.id": institution_id, "types.type": relation_type}},
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [{"$project": {"source": 1}}],
            }
        },
        {"$unwind": "$works"},
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "sources",
                "localField": "works.source.id",
                "foreignField": "_id",
                "as": "source",
                "pipeline": [{"$project": {"apc": 1}}],
            }
        },
        {"$unwind": "$source"},
        {"$project": {"_id": 0, "source": 1, "names": 1}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_departments_apc_expenses_by_faculty(affiliation_id: str, query_params: QueryParams) -> CommandCursor:
    return get_affiliations_apc_expenses_by_institution(affiliation_id, "department", query_params)


def get_groups_apc_expenses_by_faculty_or_department(affiliation_id: str, query_params: QueryParams) -> CommandCursor:
    institution_id = (
        database["affiliations"]
        .aggregate(
            [
                {"$match": {"_id": affiliation_id}},
                {"$unwind": "$relations"},
                {"$match": {"relations.types.type": "education"}},
            ]
        )
        .next()
        .get("relations", {})
        .get("id", None)
    )
    pipeline = [
        {
            "$match": {
                "relations.id": institution_id,
                "types.type": "group",
            }
        },
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [
                    {"$match": {"authors.affiliations.id": affiliation_id}},
                    {"$project": {"source": 1, "authors": 1}},
                ],
            }
        },
        {"$unwind": "$works"},
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "sources",
                "localField": "works.source.id",
                "foreignField": "_id",
                "as": "source",
                "pipeline": [{"$project": {"apc": 1}}],
            }
        },
        {"$unwind": "$source"},
        {"$project": {"_id": 0, "source": 1, "names": 1}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_affiliations_works_citations_count_by_institution(
    institution_id: str, relation_type: str, query_params: QueryParams
) -> CommandCursor:
    pipeline = [
        {"$match": {"relations.id": institution_id, "types.type": relation_type}},
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [
                    {
                        "$addFields": {
                            "scholar_citations_count": {
                                "$ifNull": [
                                    {
                                        "$arrayElemAt": [
                                            {
                                                "$map": {
                                                    "input": {
                                                        "$filter": {
                                                            "input": "$citations_count",
                                                            "as": "citation",
                                                            "cond": {
                                                                "$eq": [
                                                                    "$$citation.source",
                                                                    "scholar",
                                                                ]
                                                            },
                                                        }
                                                    },
                                                    "as": "filtered",
                                                    "in": "$$filtered.count",
                                                }
                                            },
                                            0,
                                        ]
                                    },
                                    0,
                                ]
                            }
                        }
                    },
                    {"$project": {"scholar_citations_count": 1}},
                ],
            }
        },
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {"$project": {"_id": 0, "works": 1, "name": {"$first": "$names.name"}}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_departments_works_citations_count_by_faculty(affiliation_id: str, query_params: QueryParams) -> CommandCursor:
    return get_affiliations_works_citations_count_by_institution(affiliation_id, "department", query_params)


def get_groups_works_citations_count_by_faculty_or_department(
    affiliation_id: str, query_params: QueryParams
) -> CommandCursor:
    institution_id = (
        database["affiliations"]
        .aggregate(
            [
                {"$match": {"_id": affiliation_id}},
                {"$unwind": "$relations"},
                {"$match": {"relations.types.type": "education"}},
            ]
        )
        .next()
        .get("relations", {})
        .get("id", None)
    )
    pipeline = [
        {
            "$match": {
                "relations.id": institution_id,
                "types.type": "group",
            }
        },
        {
            "$lookup": {
                "from": "works",
                "localField": "_id",
                "foreignField": "authors.affiliations.id",
                "as": "works",
                "pipeline": [
                    {"$match": {"authors.affiliations.id": affiliation_id}},
                    {
                        "$addFields": {
                            "scholar_citations_count": {
                                "$ifNull": [
                                    {
                                        "$arrayElemAt": [
                                            {
                                                "$map": {
                                                    "input": {
                                                        "$filter": {
                                                            "input": "$citations_count",
                                                            "as": "citation",
                                                            "cond": {
                                                                "$eq": [
                                                                    "$$citation.source",
                                                                    "scholar",
                                                                ]
                                                            },
                                                        }
                                                    },
                                                    "as": "filtered",
                                                    "in": "$$filtered.count",
                                                }
                                            },
                                            0,
                                        ]
                                    },
                                    0,
                                ]
                            }
                        }
                    },
                    {"$project": {"scholar_citations_count": 1}},
                ],
            }
        },
    ]
    set_plot_product_filters(pipeline, query_params)
    pipeline += [
        {"$project": {"_id": 0, "works": 1, "name": {"$first": "$names.name"}}},
    ]
    return database["affiliations"].aggregate(pipeline)


def get_active_authors_by_sex(affiliation_id: str) -> CommandCursor:
    pipeline = [
        {
            "$match": {
                "affiliations": {"$elemMatch": {"id": affiliation_id, "end_date": -1}},
                "ranking.source": "staff",
            }
        },
        {"$project": {"_id": 0, "sex": 1}},
    ]
    return database["person"].aggregate(pipeline)


def get_active_authors_by_age_range(affiliation_id: str) -> CommandCursor:
    pipeline = [
        {
            "$match": {
                "affiliations": {"$elemMatch": {"id": affiliation_id, "end_date": -1}},
                "ranking.source": "staff",
            }
        },
        {"$project": {"_id": 0, "birthdate": 1}},
    ]
    return database["person"].aggregate(pipeline)


def get_products_by_author_age_and_person(person_id: str, query_params: QueryParams) -> CommandCursor:
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {"$project": {"authors": 1, "date_published": 1, "year_published": 1}},  # type: ignore
        {
            "$lookup": {
                "from": "person",  # type: ignore
                "localField": "authors.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "pipeline": [{"$project": {"birthdate": 1}}],  # type: ignore
                "as": "author",  # type: ignore
            }
        },
        {"$unwind": "$author"},  # type: ignore
        {
            "$project": {
                "work.date_published": "$date_published",  # type: ignore
                "work.year_published": "$year_published",  # type: ignore
                "birthdate": "$author.birthdate",  # type: ignore
            }
        },
    ]
    return database["works"].aggregate(pipeline)


def get_coauthorship_by_country_map_by_affiliation(affiliation_id: str, query_params: QueryParams) -> list:
    data = []
    pipeline = [
        {"$match": {"authors.affiliations.id": affiliation_id}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {"$unwind": "$authors"},  # type: ignore
        {"$unwind": "$authors.affiliations"},  # type: ignore
        {"$group": {"_id": "$authors.affiliations.id", "count": {"$sum": 1}}},  # type: ignore
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "_id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliation",  # type: ignore
                "pipeline": [  # type: ignore
                    {
                        "$project": {
                            "addresses.country_code": 1,
                            "addresses.country": 1,
                        }
                    }
                ],
            }
        },
        {"$unwind": "$affiliation"},  # type: ignore
        {"$unwind": "$affiliation.addresses"},  # type: ignore
    ]
    for work in database["works"].aggregate(pipeline):
        data.append(work)
    return data


def get_coauthorship_by_country_map_by_person(person_id: str, query_params: QueryParams) -> list:
    data = []
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {"$unwind": "$authors"},  # type: ignore
        {"$unwind": "$authors.affiliations"},  # type: ignore
        {"$group": {"_id": "$authors.affiliations.id", "count": {"$sum": 1}}},  # type: ignore
        {"$unwind": "$_id"},  # type: ignore
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "_id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliation",  # type: ignore
                "pipeline": [  # type: ignore
                    {
                        "$project": {
                            "addresses.country_code": 1,
                            "addresses.country": 1,
                        }
                    }
                ],
            }
        },
        {
            "$project": {
                "count": 1,  # type: ignore
                "affiliation.addresses.country_code": 1,  # type: ignore
                "affiliation.addresses.country": 1,  # type: ignore
            }
        },
        {"$unwind": "$affiliation"},  # type: ignore
        {"$unwind": "$affiliation.addresses"},  # type: ignore
    ]
    for work in database["works"].aggregate(pipeline):
        data.append(work)
    return data


def get_coauthorship_by_colombian_department_map_by_affiliation(affiliation_id: str, query_params: QueryParams) -> list:
    data = []
    pipeline = [
        {"$match": {"authors.affiliations.id": affiliation_id}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {"$unwind": "$authors"},  # type: ignore
        {"$group": {"_id": "$authors.affiliations.id", "count": {"$sum": 1}}},  # type: ignore
        {"$unwind": "$_id"},  # type: ignore
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "_id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliation",  # type: ignore
                "pipeline": [  # type: ignore
                    {
                        "$project": {
                            "addresses.country_code": 1,
                            "addresses.city": 1,
                        }
                    }
                ],
            }
        },
        {"$unwind": "$affiliation"},  # type: ignore
        {"$unwind": "$affiliation.addresses"},  # type: ignore
    ]
    for work in database["works"].aggregate(pipeline):
        data.append(work)
    return data


def get_coauthorship_by_colombian_department_map_by_person(person_id: str, query_params: QueryParams) -> list:
    data = []
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {"$unwind": "$authors"},  # type: ignore
        {"$group": {"_id": "$authors.affiliations.id", "count": {"$sum": 1}}},  # type: ignore
        {"$unwind": "$_id"},  # type: ignore
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "_id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliation",  # type: ignore
            }
        },
        {
            "$project": {
                "count": 1,  # type: ignore
                "affiliation.addresses.country_code": 1,  # type: ignore
                "affiliation.addresses.city": 1,  # type: ignore
            }
        },
        {"$unwind": "$affiliation"},  # type: ignore
        {"$unwind": "$affiliation.addresses"},  # type: ignore
    ]
    for work in database["works"].aggregate(pipeline):
        data.append(work)
    return data


def get_collaboration_network(affiliation_id: str) -> CommandCursor:
    pipeline = [
        {"$match": {"_id": affiliation_id}},
        {"$project": {"coauthorship_network": 1}},
        {
            "$lookup": {
                "from": "affiliations_edges",
                "localField": "_id",
                "foreignField": "_id",
                "as": "complement",
            }
        },
        {"$unwind": "$complement"},
        {
            "$project": {
                "coauthorship_network": {
                    "nodes": "$coauthorship_network.nodes",
                    "edges": {
                        "$concatArrays": [
                            "$coauthorship_network.edges",
                            "$complement.coauthorship_network.edges",
                        ]
                    },
                }
            }
        },
    ]
    return calculations_database["affiliations"].aggregate(pipeline)


def get_works_rankings_by_person(person_id: str, query_params: QueryParams) -> Tuple[Generator, int]:
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "sources",  # type: ignore
                "localField": "source.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "source_data",  # type: ignore
                "pipeline": [  # type: ignore
                    {"$project": {"_id": 1, "ranking": 1}},
                ],
            }
        },
        {"$unwind": "$source_data"},  # type: ignore
        {"$project": {"_id": 1, "source_data": 1, "date_published": 1}},  # type: ignore
    ]
    count_pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(count_pipeline, query_params)
    count_pipeline += [
        {"$count": "total_results"},  # type: ignore
    ]
    total_results = next(database["works"].aggregate(count_pipeline), {"total_results": 0})["total_results"]
    works = database["works"].aggregate(pipeline)
    return work_generator.get(works), total_results


def get_products_by_database_by_affiliation(affiliation_id: str) -> dict:
    return {
        "minciencias": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                ]
            }
        ),
        "openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "openalex"},
                ]
            }
        ),
        "scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "scienti": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_minciencias": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "openalex"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "minciencias_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                ]
            }
        ),
        "minciencias_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "scienti_minciencias_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_minciencias_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "minciencias_openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "minciencias_openalex_scholar_scienti": database["works"].count_documents(
            {
                "$and": [
                    {"authors.affiliations.id": affiliation_id},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
    }


def get_products_by_database_by_person(person_id: str) -> dict:
    return {
        "minciencias": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                ]
            }
        ),
        "openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "openalex"},
                ]
            }
        ),
        "scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "scienti": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_minciencias": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "openalex"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "minciencias_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                ]
            }
        ),
        "minciencias_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "scienti_minciencias_openalex": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_minciencias_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "scienti_openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
        "minciencias_openalex_scholar": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                ]
            }
        ),
        "minciencias_openalex_scholar_scienti": database["works"].count_documents(
            {
                "$and": [
                    {"authors.id": ObjectId(person_id)},
                    {"updated.source": "minciencias"},
                    {"updated.source": "openalex"},
                    {"updated.source": "scholar"},
                    {"updated.source": "scienti"},
                ]
            }
        ),
    }


def set_plot_product_filters(pipeline: list, query_params: QueryParams) -> None:
    set_plot_product_type_filters(pipeline, query_params.product_types)
    set_plot_year_filters(pipeline, query_params.years)
    set_plot_status_filters(pipeline, query_params.status)
    set_plot_subject_filters(pipeline, query_params.subjects)
    set_plot_country_filters(pipeline, query_params.countries)
    set_plot_groups_ranking_filters(pipeline, query_params.groups_ranking)
    set_plot_authors_ranking_filters(pipeline, query_params.authors_ranking)


def set_plot_product_type_filters(pipeline: list, type_filters: str | None) -> None:
    if not type_filters:
        return
    match_filters = []
    for type_filter in type_filters.split(","):
        source, type_name = type_filter.split("_")
        match_filters.append({"works.types.source": source, "works.types.type": type_name})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_plot_year_filters(pipeline: list, years: str | None) -> None:
    if not years:
        return
    year_list = [int(year) for year in years.split(",")]
    first_year = min(year_list)
    last_year = max(year_list)
    pipeline += [{"$match": {"works.year_published": {"$gte": first_year, "$lte": last_year}}}]


def set_plot_status_filters(pipeline: list, status: str | None) -> None:
    if not status:
        return
    match_filters = []
    for single_status in status.split(","):
        if single_status == "unknown":
            match_filters.append({"works.open_access.open_access_status": None})
        elif single_status == "open":
            match_filters.append({"works.open_access.open_access_status": {"$nin": [None, "closed"]}})  # type: ignore
        else:
            match_filters.append({"works.open_access.open_access_status": single_status})  # type: ignore
    pipeline += [{"$match": {"$or": match_filters}}]


def set_plot_subject_filters(pipeline: list, subjects: str | None) -> None:
    if not subjects:
        return
    match_filters = []
    for subject in subjects.split(","):
        params = subject.split("_")
        if len(params) == 1:
            return
        match_filters.append({"works.subjects.subjects": {"$elemMatch": {"level": int(params[0]), "name": params[1]}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_plot_country_filters(pipeline: list, countries: str | None) -> None:
    if not countries:
        return
    match_filters = []
    for country in countries.split(","):
        match_filters.append({"works.authors.affiliations": {"$elemMatch": {"country_code": country}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_plot_groups_ranking_filters(pipeline: list, groups_ranking: str | None) -> None:
    if not groups_ranking:
        return
    match_filters = []
    for ranking in groups_ranking.split(","):
        match_filters.append({"works.groups": {"$elemMatch": {"ranking": ranking}}})
    pipeline += [{"$match": {"$or": match_filters}}]


def set_plot_authors_ranking_filters(pipeline: list, authors_ranking: str | None) -> None:
    if not authors_ranking:
        return
    match_filters = []
    for ranking in authors_ranking.split(","):
        match_filters.append({"works.authors": {"$elemMatch": {"ranking": ranking}}})
    pipeline += [{"$match": {"$or": match_filters}}]



from typing import Generator

from bson import ObjectId

from domain.models.base_model import QueryParams
from infrastructure.generators import work_generator
from infrastructure.mongo import database
from infrastructure.repositories import work_repository


def get_works_csv_by_person(person_id: str, query_params: QueryParams) -> Generator:
    pipeline = [
        {"$match": {"authors.id": ObjectId(person_id)}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "authors.affiliations.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliations_data",  # type: ignore
                "pipeline": [{"$project": {"id": "$_id", "addresses.country": 1, "ranking": 1}}],  # type: ignore
            }
        },
        {
            "$lookup": {
                "from": "sources",  # type: ignore
                "localField": "source.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "source_data",  # type: ignore
                "pipeline": [  # type: ignore
                    {
                        "$project": {
                            "external_urls": 1,
                            "ranking": 1,
                            "publisher": 1,
                            "apc": 1,
                        }
                    }
                ],
            }
        },
        {
            "$project": {
                "external_ids": 1,  # type: ignore
                "authors": 1,  # type: ignore
                "affiliations_data": 1,  # type: ignore
                "bibliographic_info": 1,  # type: ignore
                "citations_count": 1,  # type: ignore
                "open_access": 1,  # type: ignore
                "subjects": 1,  # type: ignore
                "titles": 1,  # type: ignore
                "types": 1,  # type: ignore
                "source": 1,  # type: ignore
                "source_data": 1,  # type: ignore
                "year_published": 1,  # type: ignore
                "ranking": 1,  # type: ignore
                "abstract": 1,  # type: ignore
            }
        },
    ]
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)


def get_works_csv_by_affiliation(affiliation_id: str, query_params: QueryParams) -> Generator:
    pipeline = [
        {"$match": {"authors.affiliations.id": affiliation_id}},
    ]
    work_repository.set_product_filters(pipeline, query_params)
    pipeline += [
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "authors.affiliations.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliations_data",  # type: ignore
                "pipeline": [{"$project": {"id": "$_id", "addresses.country": 1, "ranking": 1}}],  # type: ignore
            }
        },
        {
            "$lookup": {
                "from": "sources",  # type: ignore
                "localField": "source.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "source_data",  # type: ignore
                "pipeline": [  # type: ignore
                    {
                        "$project": {
                            "external_urls": 1,
                            "ranking": 1,
                            "publisher": 1,
                            "apc": 1,
                        }
                    }
                ],
            }
        },
        {
            "$project": {
                "external_ids": 1,  # type: ignore
                "authors": 1,  # type: ignore
                "affiliations_data": 1,  # type: ignore
                "bibliographic_info": 1,  # type: ignore
                "open_access": 1,  # type: ignore
                "citations_count": 1,  # type: ignore
                "subjects": 1,  # type: ignore
                "titles": 1,  # type: ignore
                "types": 1,  # type: ignore
                "source": 1,  # type: ignore
                "source_data": 1,  # type: ignore
                "year_published": 1,  # type: ignore
                "ranking": 1,  # type: ignore
                "abstract": 1,  # type: ignore
            }
        },
    ]
    cursor = database["works"].aggregate(pipeline)
    return work_generator.get(cursor)



from flask import Blueprint, redirect
from werkzeug import Response

router = Blueprint("apidoc", __name__)


@router.route("/apidoc")
def redirect_to_index() -> Response:
    return redirect("static/index.html")


@router.route("/docs")
def redirect_to_index_docs() -> Response:
    return redirect("static/index.html")


@router.route("/robots.txt")
def redirect_to_robots() -> Response:
    return redirect("static/robots.txt")


@router.route("/google54c758d287263571.html")
def redirect_to_google() -> Response:
    return redirect("static/google54c758d287263571.html")



def parse_products_by_database(data: dict) -> dict:
    venn_source = {
        "scienti": data.get("scienti", 0),
        "minciencias": data.get("minciencias", 0),
        "openalex": data.get("openalex", 0),
        "scholar": data.get("scholar", 0),
        "scienti_minciencias": data.get("scienti_minciencias", 0),
        "scienti_openalex": data.get("scienti_openalex", 0),
        "scienti_scholar": data.get("scienti_scholar", 0),
        "minciencias_openalex": data.get("minciencias_openalex", 0),
        "minciencias_scholar": data.get("minciencias_scholar", 0),
        "openalex_scholar": data.get("openalex_scholar", 0),
        "scienti_minciencias_openalex": data.get("scienti_minciencias_openalex", 0),
        "scienti_minciencias_scholar": data.get("scienti_minciencias_scholar", 0),
        "scienti_openalex_scholar": data.get("scienti_openalex_scholar", 0),
        "minciencias_openalex_scholar": data.get("minciencias_openalex_scholar", 0),
        "minciencias_openalex_scholar_scienti": data.get("minciencias_openalex_scholar_scienti", 0),
    }
    return {"plot": venn_source}



from functools import wraps
from datetime import datetime
from itertools import chain
from typing import Callable, Iterable, Generator
from collections import Counter, defaultdict

from currency_converter import CurrencyConverter
from pymongo.command_cursor import CommandCursor

from domain.constants.open_access_status import open_access_status_dict
from domain.models.affiliation_model import Affiliation
from domain.helpers import get_works_h_index_by_scholar_citations
from domain.models.calculations_model import Calculations


def get_percentage(func: Callable[..., list]) -> Callable[..., dict]:
    @wraps(func)
    def wrapper(*args: Iterable, **kwargs: dict) -> dict:
        data = func(*args, **kwargs)
        total = sum(item["value"] for item in data)
        for item in data:
            item["percentage"] = round(item["value"] / total * 100, 2) if total else 0
        return {"plot": data, "sum": total}

    return wrapper


@get_percentage
def parse_citations_by_affiliations(data: CommandCursor) -> list:
    plot: list = []
    for item in data:
        citations_count = item.get("citations_count", {})
        scholar_citations_count: dict = next(filter(lambda x: x["source"] == "scholar", citations_count), {})
        plot.append({"name": item.get("name", "No name"), "value": scholar_citations_count.get("count", 0)})
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_apc_expenses_by_affiliations(data: CommandCursor) -> list:
    result: defaultdict = defaultdict(int)
    currency_converter = CurrencyConverter()
    for item in data:
        apc_charges = item.get("source").get("apc").get("charges", 0)
        apc_currency = item.get("source").get("apc").get("currency", "USD")
        if apc_currency in ["IRR", "NGN"]:
            continue
        usd_charges = currency_converter.convert(apc_charges, apc_currency, "USD")
        result[item.get("names", [{"name": "No name"}])[0].get("name")] += int(usd_charges)
    plot = []
    for name, value in result.items():
        plot.append({"name": name, "value": value})
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_h_index_by_affiliation(data: CommandCursor) -> list:
    plot = []
    for item in data:
        plot.append({"name": item.get("name"), "value": get_works_h_index_by_scholar_citations(item.get("works"))})
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_articles_by_publisher(works: Generator) -> list:
    data = map(
        lambda x: (
            x.source.publisher.name
            if x.source.publisher and isinstance(x.source.publisher.name, str)
            else "Sin información"
        ),
        works,
    )
    counter = Counter(data)
    plot = []
    for name, value in counter.items():
        plot += [{"name": name, "value": value}]
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_products_by_subject(works: Generator) -> list:
    data = chain.from_iterable(
        map(
            lambda x: [sub for subject in x.subjects for sub in subject.subjects if subject.source == "openalex"],
            works,
        )
    )
    results = Counter(subject.name for subject in data)
    plot = []
    for name, value in results.items():
        plot.append({"name": name, "value": value})
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_products_by_access_route(works: Generator) -> list:
    data = map(
        lambda x: (x.open_access.open_access_status if x.open_access.open_access_status else "no_info"),
        works,
    )
    counter = Counter(data)
    plot = []
    for name, value in counter.items():
        plot.append({"name": open_access_status_dict.get(name), "value": value})
    return sorted(plot, key=lambda x: x.get("value"), reverse=True)


@get_percentage
def parse_active_authors_by_sex(persons: CommandCursor) -> list:
    result: defaultdict = defaultdict(int)
    for person in persons:
        if not person.get("sex") or person.get("sex") == "":
            result["Sin información"] += 1
            continue
        result[person.get("sex")] += 1
    plot = []
    for name, value in result.items():
        plot.append({"name": name, "value": value})
    return plot


@get_percentage
def parse_active_authors_by_age_range(persons: CommandCursor) -> list:
    ranges = {"14-26": (14, 26), "27-59": (27, 59), "60+": (60, float("inf"))}
    result = {"14-26": 0, "27-59": 0, "60+": 0, "Sin información": 0}
    for person in persons:
        if not person.get("birthdate") or person.get("birthdate") == -1:
            result["Sin información"] += 1
            continue
        birthdate = datetime.fromtimestamp(person.get("birthdate")).year
        age = datetime.now().year - birthdate
        for name, (low_age, high_age) in ranges.items():
            if low_age <= age <= high_age:
                result[name] += 1
                break
    plot = []
    for name, value in result.items():
        plot.append({"name": name, "value": value})
    return plot


@get_percentage
def parse_articles_by_scienti_category(works: list) -> list:
    total_works = len(works)
    data = filter(
        lambda x: x.source == "scienti" and x.rank and x.rank.split("_")[-1] in ["A", "A1", "B", "C", "D"],
        chain.from_iterable(map(lambda x: x.ranking, works)),
    )
    counter = Counter(map(lambda x: x.rank.split("_")[-1], data))
    plot = []
    for name, value in counter.items():
        plot.append({"name": name, "value": value})
    plot.append({"name": "Sin información", "value": total_works - sum(counter.values())})
    return plot


@get_percentage
def parse_articles_by_scimago_quartile(works: Generator) -> list:
    valid_sources = {"Scimago Best Quartile", "scimago Best Quartile"}
    data = []
    total_articles = 0
    for work in works:
        total_articles += 1
        work_date = work.date_published
        if not work_date:
            continue
        for ranking in work.source.ranking:
            if (
                ranking.source in valid_sources
                and ranking.rank != "-"
                and ranking.from_date <= work_date <= ranking.to_date
            ):
                data.append(ranking.rank)
                break
    counter = Counter(data)
    plot = [{"name": "Sin información", "value": total_articles - len(data)}]
    plot.extend({"name": name, "value": value} for name, value in counter.items())
    return plot


@get_percentage
def parse_articles_by_publishing_institution(works: Generator, institution: Affiliation) -> list:
    result = {"Misma": 0, "Diferente": 0, "Sin información": 0}
    names = []
    if institution:
        names = list(set([name.name.lower() for name in institution.names]))
    for work in works:
        if (
            not work.source.publisher
            or not work.source.publisher.name
            or not isinstance(work.source.publisher.name, str)
        ):
            result["Sin información"] += 1
            continue
        if work.source.publisher.name.lower() in names:
            result["Misma"] += 1
        else:
            result["Diferente"] += 1
    plot = []
    for name, value in result.items():
        plot.append({"name": name, "value": value})
    return plot


def parse_most_used_title_words(data: Calculations) -> dict:
    top_words = data.model_dump().get("top_words")
    if not top_words:
        return {
            "plot": [{"name": "Sin información", "value": 1, "percentage": 100}],
            "sum": 1,
        }
    return {"plot": sorted(top_words, key=lambda x: x.get("value"), reverse=True)}



articles_types_list: list[str] = [
    "article",
    "Artículo",
    "journal-article",
    "Artículos de investigación",
    "Publicado en revista especializada",
    "Corto (Resumen)",
    "Revisión (Survey)",
    "Caso clínico",
    "Otro artículo publicado",
    "Artículos",
]



import json
from typing import Tuple

from flask import Blueprint, request, Response, jsonify
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from domain.services import (
    work_service,
    affiliation_service,
    other_work_service,
    project_service,
    affiliation_plot_service,
    csv_service,
    patent_service,
)

affiliation_app_router = Blueprint("affiliation_app_router", __name__)

"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id Get affiliation by id
@apiName GetAffiliationById
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene una afiliación específica.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>", methods=["GET"])
def get_affiliation_by_id(affiliation_type: str, affiliation_id: str) -> Response | Tuple[Response, int]:
    try:
        data = affiliation_service.get_affiliation_by_id(affiliation_id, affiliation_type)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/affiliations Get related affiliations
@apiName GetAffiliationAffiliations
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene las afiliaciones relacionadas a una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/affiliations", methods=["GET"])
def get_affiliation_affiliations(affiliation_type: str, affiliation_id: str) -> Response | Tuple[Response, int]:
    try:
        data = affiliation_service.get_related_affiliations_by_affiliation(affiliation_id, affiliation_type)
        response_data = json.dumps(data, sort_keys=False)
        return Response(response_data, mimetype="application/json")
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/products Get works by affiliation
@apiName GetAffiliationResearchProducts
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene los productos bibliográficos de una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/products", methods=["GET"])
def get_affiliation_research_products(affiliation_id: str, affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        if query_params.plot:
            data = affiliation_plot_service.get_affiliation_plot(affiliation_id, affiliation_type, query_params)
            return jsonify(data)
        data = work_service.get_works_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/products/filters Get works filters by affiliation
@apiName GetAffiliationResearchProductsFilters
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene los filtros disponibles en los productos bibliográficos de una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/products/filters", methods=["GET"])
def get_affiliation_research_products_filters(
    affiliation_id: str, affiliation_type: str
) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = work_service.get_works_filters_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/products/csv Get works csv by affiliation
@apiName GetAffiliationResearchProductsCSV
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene los productos bibliográficos de una afiliación en formato CSV.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/products/csv", methods=["GET"])
def get_works_csv_by_affiliation(affiliation_type: str, affiliation_id: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = csv_service.get_works_csv_by_affiliation(affiliation_id, query_params)
        response = Response(data, content_type="text/csv")
        response.headers["Content-Disposition"] = "attachment; filename=affiliation.csv"
        return response
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/other_works Get other works by affiliation
@apiName GetAffiliationOtherWorks
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene otros productos de una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/other_works", methods=["GET"])
def get_affiliation_research_other_works(affiliation_id: str, affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = other_work_service.get_other_works_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/patents Get patents by affiliation
@apiName GetAffiliationPatents
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene los patentes de una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/patents", methods=["GET"])
def get_affiliation_research_patents(affiliation_id: str, affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = patent_service.get_patents_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/affiliations/:affiliation_type/:affiliation_id/research/projects Get projects by affiliation
@apiName GetAffiliationProjects
@apiGroup Affiliation
@apiVersion 1.0.0
@apiDescription Obtiene los proyectos de una afiliación.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.
"""


@affiliation_app_router.route("/<affiliation_type>/<affiliation_id>/research/projects", methods=["GET"])
def get_affiliation_research_projects(affiliation_id: str, affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = project_service.get_projects_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from bson import ObjectId
from pydantic import BaseModel, Field, field_validator, model_validator
from domain.models.base_model import (
    PyObjectId,
    CitationsCount,
    Updated,
    ExternalId,
    Type,
    Identifier,
    Subject,
    Ranking,
    ExternalUrl,
    BirthPlace,
)


class Affiliation(BaseModel):
    id: str | None = None
    name: str | None = None
    position: str | None = None
    start_date: int | str | None = None
    end_date: int | str | None = None
    types: list[Type] | None = Field(default_factory=list)
    external_urls: list[ExternalUrl] | None = Field(default_factory=list)

    @field_validator("start_date")
    @classmethod
    def set_start_date(cls, value: int | str) -> int | str:
        if value == "":
            return -1
        else:
            return value

    @field_validator("end_date")
    @classmethod
    def set_end_date(cls, value: int | str) -> int | str:
        if value == "":
            return -1
        else:
            return value

    class Config:
        json_encoders = {ObjectId: str}


class Degree(BaseModel):
    id: str | None
    date: int | None
    degree: str | None
    institutions: list | None = Field(default_factory=list)
    source: str | None


class Institution(BaseModel):
    id: str | None
    country_code: str | None
    country_id: str | None
    display_name: str | None
    lineage: list[str] | None = Field(default_factory=list)
    ror: str | None
    type: str | None


class RelatedWork(BaseModel):
    id: str | Identifier | None
    institutions: list[Institution] | None = Field(default_factory=list)
    provenance: str | None
    source: str | None
    year: int | None = None


class Person(BaseModel):
    id: PyObjectId = Field(alias="_id")
    affiliations: list[Affiliation] | None = Field(default_factory=list)
    aliases: list[str] | None = Field(default_factory=list)
    birthdate: int | str | None = None
    birthplace: BirthPlace | None = None
    citations_count: list[CitationsCount] | None = Field(default_factory=list)
    degrees: list[Degree] | None = Field(default_factory=list)
    external_ids: list[ExternalId] | None = Field(default_factory=list)
    first_names: list[str] | None = Field(default_factory=list)
    full_name: str | None = Field(serialization_alias="name")
    initials: str | None = None
    keywords: list | None = Field(default_factory=list)
    last_names: list[str] | None = Field(default_factory=list)
    marital_status: str | None = None
    products_count: int | None = None
    ranking: list[Ranking] | None = Field(default_factory=list)
    related_works: list[RelatedWork] | None = Field(default_factory=list)
    sex: str | None = None
    subjects: list[Subject] | None = Field(default_factory=list)
    updated: list[Updated] | None = Field(default_factory=list)

    affiliations_data: list[Affiliation] | None = None
    logo: str | None = None

    @field_validator("external_ids")
    @classmethod
    def delete_sensitive_external_ids(cls, value: list) -> list:
        return list(
            filter(
                lambda external_id: external_id.source
                not in ["Cédula de Ciudadanía", "Cédula de Extranjería", "Passport"],
                value,
            )
        )

    class Config:
        json_encoders = {ObjectId: str}

    @model_validator(mode="after")
    def get_logo(self) -> "Person":
        if self.affiliations_data:
            self.logo = next(filter(lambda x: x.source == "logo", self.affiliations_data[0].external_urls)).url  # type: ignore
        return self



def get_works_h_index_by_scholar_citations(works: list) -> int:
    citations = sorted([work.get("scholar_citations_count") for work in works], reverse=True)
    h_index = 0
    for i, cites in enumerate(citations):
        current_index = i + 1
        if cites >= current_index:
            h_index = current_index
        else:
            break
    return h_index



from domain.models.project_model import Project


def parse_search_results(projects: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "year_init",
        "year_end",
        "title",
        "external_ids",
        "external_urls",
    ]
    return [project.model_dump(include=include) for project in projects]


def parse_projects_by_entity(projects: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "year_init",
        "year_end",
        "title",
        "external_ids",
    ]
    return [project.model_dump(include=include) for project in projects]


def parse_project(project: Project) -> dict:
    return project.model_dump()



openalex_types_dict = {
    "peer-review": "Revisión por partes",
    "techreport": "Informe técnico",
    "masterthesis": "Tesis de maestría",
    "dataset": "Conjunto de datos",
    "editorial": "Editorial",
    "Publicado en revista especializada": "Publicado en revista especializada",
    "report": "Informe",
    "Artículos": "Artículos",
    "letter": "Carta",
    "Corto (resumen)": "Resumen",
    "reference-entry": "Entrada de referencia",
    "dissertation": "Disertación",
    "standard": "Estándar",
    "Artículos de investigación": "Artículos de investigación",
    "Artículo": "Artículo",
    "incollection": "En colección",
    "book": "Libro",
    "article": "Artículo",
    "Caso clínico": "Caso clínico",
    "paratext": "Paratexto",
    "misc": "Misceláneo",
    "erratum": "Errata",
    "Revisión (Survey)": "Revisión",
    "inproceedings": "En actas",
}



from typing import Generator, Tuple

from bson import ObjectId

from infrastructure.generators import person_generator
from domain.models.base_model import QueryParams
from infrastructure.repositories import base_repository
from domain.exceptions.not_entity_exception import NotEntityException
from domain.models.person_model import Person
from infrastructure.mongo import database


def get_person_by_id(person_id: str) -> Person:
    person_data = database["person"].aggregate(
        [
            {"$match": {"_id": ObjectId(person_id)}},
            {
                "$addFields": {
                    "filtered_affiliations": {
                        "$filter": {
                            "input": "$affiliations",
                            "as": "affiliation",
                            "cond": {"$eq": ["$$affiliation.end_date", -1]},
                        }
                    }
                }
            },
            {
                "$lookup": {
                    "from": "affiliations",
                    "localField": "filtered_affiliations.id",
                    "foreignField": "_id",
                    "as": "affiliations_data",
                    "pipeline": [
                        {"$match": {"external_urls.source": "logo"}},
                        {"$project": {"_id": 0, "external_urls": 1}},
                    ],
                }
            },
        ]
    )
    person_data = next(person_data, None)
    if not person_data:
        raise NotEntityException(f"The person with id {person_id} does not exist.")
    return Person(**person_data)


def get_persons_by_affiliation(affiliation_id: str) -> list:
    pipeline = [
        {"$match": {"affiliations.id": affiliation_id}},
        {"$project": {"_id": 1, "full_name": 1}},
    ]
    cursor = database["person"].aggregate(pipeline)
    return person_generator.get(cursor)


def search_persons(query_params: QueryParams, pipeline_params: dict | None = None) -> Tuple[Generator, int]:
    if pipeline_params is None:
        pipeline_params = {}
    pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    pipeline += [
        {
            "$addFields": {
                "filtered_affiliations": {
                    "$filter": {
                        "input": "$affiliations",
                        "as": "affiliation",
                        "cond": {"$eq": ["$$affiliation.end_date", -1]},
                    }
                }
            }
        },
        {
            "$lookup": {
                "from": "affiliations",  # type: ignore
                "localField": "filtered_affiliations.id",  # type: ignore
                "foreignField": "_id",  # type: ignore
                "as": "affiliations_data",  # type: ignore
                "pipeline": [  # type: ignore
                    {"$match": {"external_urls.source": "logo"}},
                    {"$project": {"_id": 0, "external_urls": 1}},
                ],
            }
        },
    ]
    base_repository.set_search_end_stages(pipeline, query_params, pipeline_params)
    persons = database["person"].aggregate(pipeline)
    count_pipeline = [{"$match": {"$text": {"$search": query_params.keywords}}}] if query_params.keywords else []
    count_pipeline += [
        {"$count": "total_results"},  # type: ignore
    ]
    total_results = next(database["person"].aggregate(count_pipeline), {"total_results": 0})["total_results"]
    return person_generator.get(persons), total_results



from infrastructure.mongo import database


def set_works_authors_affiliations_country():  # type: ignore
    pipeline = [
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "authors.affiliations.id",
                "foreignField": "_id",
                "as": "affiliations_data",
                "pipeline": [{"$project": {"_id": 1, "addresses.country": 1}}],
            }
        },
        {
            "$addFields": {
                "authors": {
                    "$map": {
                        "input": "$authors",
                        "as": "author",
                        "in": {
                            "$mergeObjects": [
                                "$$author",
                                {
                                    "affiliations": {
                                        "$map": {
                                            "input": "$$author.affiliations",
                                            "as": "affiliation",
                                            "in": {
                                                "$mergeObjects": [
                                                    "$$affiliation",
                                                    {
                                                        "country": {
                                                            "$let": {
                                                                "vars": {
                                                                    "matchedAffiliation": {
                                                                        "$arrayElemAt": [
                                                                            {
                                                                                "$filter": {
                                                                                    "input": "$affiliations_data",
                                                                                    "as": "affiliation_data",
                                                                                    "cond": {
                                                                                        "$eq": [
                                                                                            "$$affiliation.id",
                                                                                            "$$affiliation_data._id",
                                                                                        ]
                                                                                    },
                                                                                }
                                                                            },
                                                                            0,
                                                                        ]
                                                                    }
                                                                },
                                                                "in": {
                                                                    "$ifNull": [
                                                                        {
                                                                            "$arrayElemAt": [
                                                                                "$$matchedAffiliation.addresses.country",
                                                                                0,
                                                                            ]
                                                                        },
                                                                        None,
                                                                    ]
                                                                },
                                                            }
                                                        }
                                                    },
                                                ]
                                            },
                                        }
                                    }
                                },
                            ]
                        },
                    }
                }
            }
        },
        {"$project": {"affiliations_data": 0}},
        {
            "$merge": {
                "into": "works",
                "whenMatched": "merge",
                "whenNotMatched": "fail",
            }
        },
    ]
    database["works"].aggregate(pipeline)


def set_works_authors_affiliations_country_code():  # type: ignore
    pipeline = [
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "authors.affiliations.id",
                "foreignField": "_id",
                "as": "affiliations_data",
                "pipeline": [{"$project": {"_id": 1, "addresses.country_code": 1}}],
            }
        },
        {
            "$addFields": {
                "authors": {
                    "$map": {
                        "input": "$authors",
                        "as": "author",
                        "in": {
                            "$mergeObjects": [
                                "$$author",
                                {
                                    "affiliations": {
                                        "$map": {
                                            "input": "$$author.affiliations",
                                            "as": "affiliation",
                                            "in": {
                                                "$mergeObjects": [
                                                    "$$affiliation",
                                                    {
                                                        "country_code": {
                                                            "$let": {
                                                                "vars": {
                                                                    "matchedAffiliation": {
                                                                        "$arrayElemAt": [
                                                                            {
                                                                                "$filter": {
                                                                                    "input": "$affiliations_data",
                                                                                    "as": "affiliation_data",
                                                                                    "cond": {
                                                                                        "$eq": [
                                                                                            "$$affiliation.id",
                                                                                            "$$affiliation_data._id",
                                                                                        ]
                                                                                    },
                                                                                }
                                                                            },
                                                                            0,
                                                                        ]
                                                                    }
                                                                },
                                                                "in": {
                                                                    "$ifNull": [
                                                                        {
                                                                            "$arrayElemAt": [
                                                                                "$$matchedAffiliation.addresses.country_code",
                                                                                0,
                                                                            ]
                                                                        },
                                                                        None,
                                                                    ]
                                                                },
                                                            }
                                                        }
                                                    },
                                                ]
                                            },
                                        }
                                    }
                                },
                            ]
                        },
                    }
                }
            }
        },
        {"$project": {"affiliations_data": 0}},
        {
            "$merge": {
                "into": "works",
                "whenMatched": "merge",
                "whenNotMatched": "fail",
            }
        },
    ]
    database["works"].aggregate(pipeline)


def set_works_groups_ranking():  # type: ignore
    pipeline = [
        {
            "$lookup": {
                "from": "affiliations",
                "localField": "groups.id",
                "foreignField": "_id",
                "as": "groups_data",
                "pipeline": [{"$match": {"ranking.source": "minciencias"}}, {"$project": {"_id": 1, "ranking": 1}}],
            }
        },
        {
            "$addFields": {
                "groups": {
                    "$map": {
                        "input": "$groups",
                        "as": "group",
                        "in": {
                            "$mergeObjects": [
                                "$$group",
                                {
                                    "ranking": {
                                        "$let": {
                                            "vars": {
                                                "matchedGroup": {
                                                    "$arrayElemAt": [
                                                        {
                                                            "$filter": {
                                                                "input": "$groups_data",
                                                                "as": "group_data",
                                                                "cond": {
                                                                    "$eq": [
                                                                        "$$group.id",
                                                                        "$$group_data._id",
                                                                    ]
                                                                },
                                                            }
                                                        },
                                                        0,
                                                    ]
                                                }
                                            },
                                            "in": {
                                                "$ifNull": [
                                                    {
                                                        "$arrayElemAt": [
                                                            {
                                                                "$map": {
                                                                    "input": {
                                                                        "$filter": {
                                                                            "input": "$$matchedGroup.ranking",
                                                                            "as": "rankData",
                                                                            "cond": {
                                                                                "$eq": [
                                                                                    "$$rankData.source",
                                                                                    "minciencias",
                                                                                ]
                                                                            },
                                                                        }
                                                                    },
                                                                    "as": "filteredRank",
                                                                    "in": "$$filteredRank.rank",
                                                                }
                                                            },
                                                            0,
                                                        ]
                                                    },
                                                    None,
                                                ]
                                            },
                                        }
                                    }
                                },
                            ]
                        },
                    }
                }
            }
        },
        {"$project": {"groups_data": 0}},
        {
            "$merge": {
                "into": "works",
                "whenMatched": "merge",
                "whenNotMatched": "fail",
            }
        },
    ]
    database["works"].aggregate(pipeline)


def set_works_authors_ranking():  # type: ignore
    pipeline = [
        {
            "$lookup": {
                "from": "person",
                "localField": "authors.id",
                "foreignField": "_id",
                "as": "authors_data",
                "pipeline": [{"$match": {"ranking.source": "minciencias"}}, {"$project": {"_id": 1, "ranking": 1}}],
            }
        },
        {
            "$addFields": {
                "authors": {
                    "$map": {
                        "input": "$authors",
                        "as": "author",
                        "in": {
                            "$mergeObjects": [
                                "$$author",
                                {
                                    "ranking": {
                                        "$let": {
                                            "vars": {
                                                "matchedAuthor": {
                                                    "$arrayElemAt": [
                                                        {
                                                            "$filter": {
                                                                "input": "$authors_data",
                                                                "as": "author_data",
                                                                "cond": {
                                                                    "$eq": [
                                                                        "$$author.id",
                                                                        "$$author_data._id",
                                                                    ]
                                                                },
                                                            }
                                                        },
                                                        0,
                                                    ]
                                                }
                                            },
                                            "in": {
                                                "$ifNull": [
                                                    {
                                                        "$arrayElemAt": [
                                                            {
                                                                "$map": {
                                                                    "input": {
                                                                        "$filter": {
                                                                            "input": "$$matchedAuthor.ranking",
                                                                            "as": "rankData",
                                                                            "cond": {
                                                                                "$eq": [
                                                                                    "$$rankData.source",
                                                                                    "minciencias",
                                                                                ]
                                                                            },
                                                                        }
                                                                    },
                                                                    "as": "filteredRank",
                                                                    "in": "$$filteredRank.rank",
                                                                }
                                                            },
                                                            0,
                                                        ]
                                                    },
                                                    None,
                                                ]
                                            },
                                        }
                                    }
                                },
                            ]
                        },
                    }
                }
            }
        },
        {"$project": {"authors_data": 0}},
        {
            "$merge": {
                "into": "works",
                "whenMatched": "merge",
                "whenNotMatched": "fail",
            }
        },
    ]
    database["works"].aggregate(pipeline)



from domain.constants.institutions import institutions_list
from infrastructure.mongo import database


def get_last_db_update() -> int:
    pipeline = [{"$group": {"_id": None, "max_time": {"$max": "$time"}}}]
    return next(database["log"].aggregate(pipeline), {"max_time": 0}).get("max_time", 0)


def get_entity_count(entity: str, affiliation_type: str | None = None) -> int:
    if affiliation_type:
        if affiliation_type == "institution":
            return database[entity].count_documents({"types.type": {"$in": institutions_list}})
        return database[entity].count_documents({"types.type": affiliation_type})
    return database[entity].count_documents({})



from domain.models.source_model import Source
from domain.models.work_model import Work
from infrastructure.repositories import source_repository


def update_work_source(work: Work) -> None:
    if work.source.id:
        source = source_repository.get_source_by_id(work.source.id)
        set_serials(work, source)
        set_scimago_quartile(work, source)


def update_csv_work_source(work: Work) -> None:
    if work.source.id:
        source_data = work.source_data[0]
        set_csv_scimago_quartile(work, source_data)
        set_source_urls(work, source_data)
        if source_data.publisher:
            work.publisher = str(source_data.publisher.name)
        if source_data.apc.charges and source_data.apc.currency:
            work.source_apc = str(source_data.apc.charges) + " / " + str(source_data.apc.currency)
        work.source_name = work.source.name


def set_source_urls(work: Work, source: Source) -> None:
    source_urls = []
    for external_url in source.external_urls:
        source_urls.append(str(external_url.url))
    work.source_urls = " | ".join(set(source_urls))


def set_scimago_quartile(work: Work, source: Source) -> None:
    if source.ranking:
        for ranking in source.ranking:
            condition = (
                ranking.source == "scimago Best Quartile"
                and ranking.rank != "-"
                and work.date_published
                and ranking.from_date <= work.date_published <= ranking.to_date
            )
            if condition:
                work.source.scimago_quartile = ranking.rank
                break


def set_csv_scimago_quartile(work: Work, source: Source) -> None:
    for ranking in source.ranking:
        condition = (
            ranking.source == "scimago Best Quartile"
            and ranking.rank != "-"
            and work.date_published
            and ranking.from_date <= work.date_published <= ranking.to_date
        )
        if condition:
            work.scimago_quartile = ranking.rank
            break


def set_serials(work: Work, source: Source) -> None:
    if source.external_ids:
        serials = {}
        for external_id in source.external_ids:
            serials[external_id.source] = external_id.id
        work.source.serials = serials



institutions_list: list[str] = [
    "Archive",
    "Company",
    "Education",
    "Facility",
    "Government",
    "Healthcare",
    "Nonprofit",
    "Other",
    "archive",
    "company",
    "education",
    "facility",
    "government",
    "healthcare",
    "nonprofit",
    "other",
    "institution",
]



import json
import os
from collections import defaultdict
from math import log

import pandas as pd


def parse_coauthorship_by_country_map(data: list) -> dict:
    countries: defaultdict = defaultdict(lambda: {"count": 0, "name": ""})
    for item in data:
        addresses = item.get("affiliation", {}).get("addresses", {})
        country_code = addresses.get("country_code")
        country_name = addresses.get("country")
        if country_code and country_name:
            country_data = countries[country_code]
            country_data["count"] += item["count"]
            country_data["name"] = country_name
    for country_data in countries.values():
        country_data["log_count"] = log(country_data["count"])  # type: ignore
    worldmap_path = os.path.join(os.path.dirname(__file__), "concerns/worldmap.json")
    with open(worldmap_path, "r") as worldmap_file:
        plot = json.load(worldmap_file)
    for feature in plot["features"]:
        country_code = feature["properties"].get("country_code")
        if country_code in countries:
            country_data = countries[country_code]
            feature["properties"]["count"] = country_data["count"]
            feature["properties"]["log_count"] = country_data["log_count"]
        else:
            feature["properties"]["count"] = 0
            feature["properties"]["log_count"] = 0
    return {"plot": plot}


def get_coauthorship_by_colombian_department_map(data: list) -> dict:
    cities_by_state_path = os.path.join(os.path.dirname(__file__), "concerns/cities_by_state.csv")
    cities_by_state = pd.read_csv(cities_by_state_path)
    city_to_state = dict(zip(cities_by_state["MUNICIPIO"], cities_by_state["DEPARTAMENTO"]))
    states = {}
    for item in data:
        addresses = item.get("affiliation", {}).get("addresses", {})
        if addresses.get("country_code") and addresses.get("city"):
            city = addresses["city"]
            state = city_to_state.get(city)
            if state:
                if state not in states:
                    states[state] = {"count": 0, "name": state}
                states[state]["count"] += item["count"]
    for state_data in states.values():
        state_data["log_count"] = log(state_data["count"])
    colombiamap_path = os.path.join(os.path.dirname(__file__), "concerns/colombiamap.json")
    with open(colombiamap_path, "r") as colombiamap_file:
        plot = json.load(colombiamap_file)
    for feature in plot["features"]:
        state = feature["properties"]["NOMBRE_DPT"].capitalize()
        if "bogota" in state.lower():
            state = "Bogotá D.C."
        state_data = states.get(state, {"count": 0, "log_count": 0})
        feature["properties"]["count"] = state_data["count"]
        feature["properties"]["log_count"] = state_data["log_count"]
    return {"plot": plot}



from flask import Blueprint

from config import settings

ping_router = Blueprint("ping_router", __name__)


@ping_router.route("/ping", methods=["GET"])
def read_root() -> dict:
    """Ping the API."""
    result = {"ping": str(settings.MONGO_URI)}
    return result



from typing import Tuple

from flask import Blueprint, jsonify, request, Response
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from domain.services import api_expert_service

affiliation_api_router = Blueprint("affiliation_api_router", __name__)


"""
@api {get} /affiliation/:affiliation_type/:affiliation_id/research/products Get works by affiliation
@apiName GetWorksByAffiliation
@apiGroup API Expert
@apiVersion 1.0.0
@apiDescription Obtiene los productos bibliográficos asociados a una afiliación específica.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
@apiParam {String} affiliation_id ID de la afiliación.

@apiSuccessExample {json} Respuesta exitosa:
    HTTP/1.1 200 OK
    {
    "data": [
    {
      "apc": {
        "paid": {
          "currency": "EUR",
          "provenance": "openalex",
          "source": "doaj",
          "value": 3690,
          "value_usd": 4790
        }
      },
      "authors": [
        {
          "affiliations": [
            {
              "end_date": -1,
              "geo": {
                "city": "Darlinghurst",
                "country": "Australia",
                "country_code": "AU",
                "latitude": -33.878359,
                "longitude": 151.221653,
                "region": "New South Wales"
              },
              "id": "66f6f3ac99b6ea475f3a499e",
              "name": "Garvan Institute of Medical Research",
              "ror": "https://ror.org/01b3dvp57",
              "start_date": -1,
              "types": [
                {
                  "source": "ror",
                  "type": "Nonprofit"
                },
                {
                  "source": "openalex",
                  "type": "nonprofit"
                }
              ]
            },
            {
              "end_date": -1,
              "geo": {},
              "id": "66f6f3df99b6ea475f3bcc04",
              "name": "Universidad de Nueva Gales del Sur",
              "start_date": -1,
              "types": [
                {
                  "source": "ror",
                  "type": "Education"
                },
                {
                  "source": "openalex",
                  "type": "education"
                }
              ]
            }
          ],
          "countries": [
            "Australia"
          ],
          "external_ids": [
            {
              "id": "https://openalex.org/A5077913296",
              "provenance": "openalex",
              "source": "openalex"
            },
            {
              "id": "https://orcid.org/0000-0002-5360-5180",
              "provenance": "openalex",
              "source": "orcid"
            }
          ],
          "first_names": [],
          "full_name": "Stuart G. Tangye",
          "id": "66f7c83a2046a15c590070a1",
          "last_names": [],
          "ranking": [],
          "sex": ""
        }
      ],
      "authors_count": 17,
      "bibliographic_info": {
        "end_page": "64",
        "issue": "1",
        "start_page": "24",
        "volume": "40"
      },
      "citations": [],
      "citations_by_year": [
        {
          "cited_by_count": 47,
          "year": 2024
        },
      ],
      "citations_count": [
        {
          "count": 958,
          "source": "openalex"
        },
        {
          "count": 1154,
          "provenance": "scholar",
          "source": "scholar"
        }
      ],
      "date_published": 1577854800,
      "doi": "https://doi.org/10.1007/s10875-019-00737-x",
      "external_ids": [
        {
          "id": "https://openalex.org/W3000301584",
          "provenance": "openalex",
          "source": "openalex"
        },
      ],
      "external_urls": [
        {
          "provenance": "openalex",
          "source": "open_access",
          "url": "https://link.springer.com/content/pdf/10.1007/s10875-019-00737-x.pdf"
        },
      ],
      "groups": [
        {
          "id": "66f7001199b6ea475f3be61f",
          "name": "Errores Innatos de la Inmunidad (Inmunodeficiencias Primarias)"
        }
      ],
      "id": "66fb7fd5cbfde71a09a2d66f",
      "keywords": [],
      "open_access": {
        "has_repository_fulltext": true,
        "is_open_access": true,
        "open_access_status": "hybrid",
        "url": "https://link.springer.com/content/pdf/10.1007/s10875-019-00737-x.pdf"
      },
      "ranking": [],
      "references": [],
      "scholar_citations_count": 1154,
      "source": {
        "id": "66f6eb6a99b6ea475f357b0d",
        "is_in_doaj": false,
        "name": "Journal of Clinical Immunology",
        "types": [
          {
            "source": "scimago",
            "type": "journal"
          },
        ]
      },
      "subjects": [
        {
          "source": "openalex",
          "subjects": [
            {
              "id": "66f6f03d99b6ea475f393a18",
              "level": 3,
              "name": "Immunity"
            },
          ]
        }
      ],
      "titles": [
        {
          "lang": "en",
          "source": "openalex",
          "title": "Human Inborn Errors of Immunity: 2019 Update on the Classification from the International Union of Immunological Societies Expert Committee"
        },
      ],
      "types": [
        {
          "code": "111",
          "level": 2,
          "provenance": "scienti",
          "source": "scienti",
          "type": "Publicado en revista especializada"
        },
      ],
      "updated": [
        {
          "source": "openalex",
          "time": 1727758293
        },
      ],
      "year_published": 2020
    }
    ]
    }
@apiError {String} error Mensaje de error en caso de fallo.
@apiErrorExample {json} Error en la solicitud:
    HTTP/1.1 400 Bad Request
    {
      "error": "Invalid affiliation ID"
    }
"""


@affiliation_api_router.route("/<affiliation_type>/<affiliation_id>/research/products", methods=["GET"])
def get_works_by_affiliation_api_expert(affiliation_id: str, affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = api_expert_service.get_works_by_affiliation(affiliation_id, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from bson import ObjectId
from pydantic import BaseModel, Field, model_validator, field_validator

from domain.models.base_model import (
    Type,
    Updated,
    ExternalUrl,
    CitationsCount,
    Name,
    ExternalId,
    Subject,
    Ranking,
    Address,
)


class DescriptionText(BaseModel):
    TXT_ESTADO_ARTE: str | None
    TXT_OBJETIVOS: str | None
    TXT_PLAN_TRABAJO: str | None
    TXT_PROD_DESTACADA: str | None
    TXT_RETOS: str | None
    TXT_VISION: str | None


class Description(BaseModel):
    description: DescriptionText
    source: str


class Relation(BaseModel):
    id: str | None = None
    name: str | Name | None = None
    types: list[Type] | None = None
    external_urls: list[ExternalUrl] | None = None

    @field_validator("name")
    @classmethod
    def set_name(cls, value: str) -> str:
        if isinstance(value, Name):
            return value.name
        return value


class Status(BaseModel):
    source: str | None
    status: str | None


class Affiliation(BaseModel):
    id: str = Field(alias="_id")
    abbreviations: list[str] | None = None
    addresses: list[Address] | Address | None = None
    aliases: list[str] | None = None
    birthdate: str | int | None = None
    citations_count: list[CitationsCount] | None = None
    description: list[Description] | None = None
    external_ids: list[ExternalId] | None = None
    external_urls: list[ExternalUrl] | None = None
    names: list[Name] | None = None
    products_count: int | None = None
    ranking: list[Ranking] | None = None
    relations: list[Relation] | Relation | dict | None = None
    status: list[str] | list[Status] | None = None
    subjects: list[Subject] | None = None
    types: list[Type] | None = None
    updated: list[Updated] | None = None
    year_established: int | None = None

    name: str | None = None
    logo: str | None = None
    affiliations: list[dict | Relation] | None = None
    relations_data: list[Relation] | None = None

    @model_validator(mode="after")
    def get_name(self):  # type: ignore
        es_name = next(filter(lambda x: x.lang == "es", self.names), None)
        self.name = es_name.name if es_name else self.names[0].name
        return self

    class Config:
        json_encoders = {ObjectId: str}



from typing import Tuple

from flask import Blueprint, jsonify, Response
from sentry_sdk import capture_exception

from domain.services import project_service

project_app_router = Blueprint("project_app_router", __name__)

"""
@api {get} /app/project/:project_id Get project by id
@apiName GetProjectById
@apiGroup Project
@apiVersion 1.0.0
@apiDescription Obtiene un proyecto por ID.

@apiParam {String} project_id ID del proyecto.
"""


@project_app_router.route("/<project_id>", methods=["GET"])
def get_project_by_id(project_id: str) -> Response | Tuple[Response, int]:
    try:
        data = project_service.get_project_by_id(project_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/project/:project_id/authors Get project authors
@apiName GetProjectAuthors
@apiGroup Project
@apiVersion 1.0.0
@apiDescription Obtiene los autores de un proyecto.

@apiParam {String} project_id ID del proyecto.
"""


@project_app_router.route("/<project_id>/authors", methods=["GET"])
def get_project_authors(project_id: str) -> Response | Tuple[Response, int]:
    try:
        data = project_service.get_project_authors(project_id)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from domain.models.base_model import QueryParams


def set_search_end_stages(pipeline: list, query_params: QueryParams, pipeline_params: dict | None = None) -> list:
    if pipeline_params is None:
        pipeline_params = {}
    set_sort(query_params.sort, pipeline)
    set_project(pipeline, pipeline_params.get("project"))
    set_pagination(pipeline, query_params)
    return pipeline


def set_pagination(pipeline: list, query_params: QueryParams) -> None:
    if (page := query_params.page) and (limit := query_params.limit):
        skip = (page - 1) * limit
        pipeline += [{"$skip": skip}, {"$limit": limit}]


def set_match(pipeline: list, match: dict | None) -> None:
    if not match:
        return
    pipeline += [{"$match": match}]


def set_project(pipeline: list, project: list | None) -> None:
    if not project:
        return
    pipeline += [{"$project": {"_id": 1, **{p: 1 for p in project}}}]


def set_sort(sort: str | None, pipeline: list) -> None:
    if not sort:
        return
    sort_field, direction_str = sort.split("_")
    direction = -1 if direction_str == "desc" else 1
    if sort_field == "citations":
        pipeline += [
            {
                "$addFields": {
                    "scholar_citations_count": {
                        "$ifNull": [
                            {
                                "$arrayElemAt": [
                                    {
                                        "$map": {
                                            "input": {
                                                "$filter": {
                                                    "input": "$citations_count",
                                                    "as": "citation",
                                                    "cond": {
                                                        "$eq": [
                                                            "$$citation.source",
                                                            "scholar",
                                                        ]
                                                    },
                                                }
                                            },
                                            "as": "filtered",
                                            "in": "$$filtered.count",
                                        }
                                    },
                                    0,
                                ]
                            },
                            0,
                        ]
                    }
                }
            }
        ]
        sort_field = "scholar_citations_count"
    elif sort_field == "alphabetical":
        pipeline += [
            {
                "$addFields": {
                    "titles_order": {
                        "$map": {
                            "input": "$titles",
                            "as": "title",
                            "in": {
                                "source": "$$title.source",
                                "title": "$$title.title",
                                "order": {
                                    "$switch": {
                                        "branches": [
                                            {
                                                "case": {
                                                    "$eq": [
                                                        "$$title.source",
                                                        "openalex",
                                                    ]
                                                },
                                                "then": 0,
                                            },
                                            {
                                                "case": {"$eq": ["$$title.source", "scholar"]},
                                                "then": 1,
                                            },
                                            {
                                                "case": {"$eq": ["$$title.source", "scienti"]},
                                                "then": 2,
                                            },
                                            {
                                                "case": {
                                                    "$eq": [
                                                        "$$title.source",
                                                        "minciencias",
                                                    ]
                                                },
                                                "then": 3,
                                            },
                                            {
                                                "case": {"$eq": ["$$title.source", "ranking"]},
                                                "then": 4,
                                            },
                                            {
                                                "case": {"$eq": ["$$title.source", "siiu"]},
                                                "then": 5,
                                            },
                                        ],
                                        "default": 10,
                                    }
                                },
                            },
                        }
                    }
                }
            },
            {
                "$addFields": {
                    "first_title": {
                        "$arrayElemAt": [
                            {
                                "$sortArray": {
                                    "input": "$titles_order",
                                    "sortBy": {"order": 1},
                                }
                            },
                            0,
                        ]
                    }
                }
            },
            {"$addFields": {"title": "$first_title.title"}},
            {"$unset": ["titles_order", "first_title"]},
        ]
        sort_field = "title"
    elif sort_field == "products":
        sort_field = "products_count"
    elif sort_field == "year":
        pipeline += [
            {
                "$addFields": {
                    "sort_year": {
                        "$cond": {
                            "if": {
                                "$or": [
                                    {"$eq": ["$year_published", None]},
                                    {"$eq": ["$year_published", ""]},
                                ]
                            },
                            "then": -1,
                            "else": "$year_published",
                        }
                    }
                }
            },
        ]
        sort_field = "sort_year"
    pipeline += [{"$sort": {sort_field: direction}}]



from domain.models.calculations_model import Calculations


def parse_institutional_coauthorship_network(data: Calculations) -> dict:
    network = data.coauthorship_network
    nodes = sorted(network.nodes, key=lambda x: x.degree, reverse=True)[:50]
    nodes_ids = [node.id for node in nodes]
    edges = []
    nodes = [node.model_dump() for node in nodes]
    for edge in network.edges:
        if edge.source in nodes_ids and edge.target in nodes_ids:
            edges.append(edge.model_dump())
    return {"plot": {"nodes": nodes, "edges": edges}}



from typing import Tuple

from flask import Blueprint, jsonify, Response
from sentry_sdk import capture_exception

from domain.services import (
    info_service,
)

info_app_router = Blueprint("info_app_router", __name__)


@info_app_router.route("/info", methods=["GET"])
def get_info() -> Response | Tuple[Response, int]:
    try:
        data = info_service.get_info()
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



import os
from typing import Optional
from typing_extensions import Self
from pydantic import model_validator, MongoDsn
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    class Config:
        if os.getenv("ENVIRONMENT") == "dev":
            env_file = "../.env.dev"
        elif os.getenv("ENVIRONMENT") == "prod":
            env_file = "../.env.prod"
        else:
            env_file = "../.env.local"

    environment: str = "local"

    APP_NAME: str
    APP_VERSION: str
    APP_DEBUG: bool
    APP_PORT: str | int
    APP_DOMAIN: str

    APP_URL_PREFIX: str
    API_URL_PREFIX: str

    MONGO_SERVER: str
    MONGO_USERNAME: str
    MONGO_PASSWORD: str
    MONGO_DATABASE: str
    MONGO_PORT: int
    MONGO_CALCULATIONS_DATABASE: str
    MONGO_URI: Optional[MongoDsn] = None

    SENTRY_DSN: str

    @model_validator(mode="after")
    def validate_mongo_uri(self) -> Self:
        self.MONGO_URI = MongoDsn.build(
            scheme="mongodb",
            host=self.MONGO_SERVER,
            username=self.MONGO_USERNAME,
            password=self.MONGO_PASSWORD,
            port=self.MONGO_PORT,
        )
        return self


settings: Settings = Settings()  # type: ignore



from collections import defaultdict
from typing import Generator

from currency_converter import CurrencyConverter
from pymongo.command_cursor import CommandCursor


def parse_annual_evolution_by_scienti_classification(works: Generator) -> dict:
    data: defaultdict = defaultdict(lambda: defaultdict(int))
    for work in works:
        if not work.year_published:
            continue
        for work_type in work.types:
            if work_type.source == "scienti" and work_type.level == 2:
                data[work.year_published][work_type.type] += 1
    plot = [
        {"x": year, "y": count, "type": work_type}
        for year, work_types in data.items()
        for work_type, count in work_types.items()
    ]
    return {"plot": sorted(plot, key=lambda x: (-x.get("x"), -x.get("y")))}


def parse_affiliations_by_product_type(data: CommandCursor) -> dict:
    plot = [{"x": item["name"], "y": item["works_count"], "type": item["type"]} for item in data]
    return {"plot": sorted(plot, key=lambda x: x.get("y"), reverse=True)}


def parse_annual_citation_count(works: Generator) -> dict:
    data: dict = {}
    no_info = 0
    for work in works:
        if not work.citations_by_year:
            no_info += 1
            continue
        for citation in work.citations_by_year:
            if citation.year in data.keys():
                data[citation.year] += citation.cited_by_count
            else:
                data[citation.year] = citation.cited_by_count
    plot = [{"x": year, "y": count} for year, count in sorted(data.items(), reverse=True)]
    plot += [{"x": "Sin información", "y": no_info}]
    return {"plot": plot}


def parse_annual_articles_open_access(works: Generator) -> dict:
    data: defaultdict = defaultdict(lambda: {"Abierto": 0, "Cerrado": 0, "Sin información": 0})
    for work in works:
        access_type = "Abierto" if work.open_access.is_open_access else "Cerrado"
        if work.open_access.is_open_access is None and not work.year_published:
            data["Sin año"]["Sin información"] += 1
            continue
        if work.open_access.is_open_access is None:
            data[work.year_published]["Sin información"] += 1
            continue
        if work.year_published is None:
            data["Sin año"][access_type] += 1
            continue
        data[work.year_published][access_type] += 1
    plot = [
        {"x": year, "y": count, "type": access_type}
        for year, counts in data.items()
        for access_type, count in counts.items()
    ]
    return {"plot": sorted(plot, key=lambda x: float("inf") if x.get("x") == "Sin año" else -x.get("x"))}


def parse_annual_articles_by_top_publishers(works: Generator) -> dict:
    data: defaultdict = defaultdict(lambda: defaultdict(int))
    for work in works:
        if not work.source.publisher:
            data[work.year_published]["Sin información"] += 1
            continue
        data[work.year_published][work.source.publisher.name] += 1
    plot = [
        {"x": year, "y": count, "type": publisher}
        for year, publishers in data.items()
        for publisher, count in publishers.items()
    ]
    return {"plot": sorted(plot, key=lambda x: (-x.get("x"), -x.get("y")))}


def parse_annual_apc_expenses(works: Generator) -> dict:
    data: defaultdict = defaultdict(int)
    total_apc = 0
    total_results = 0
    currency_converter = CurrencyConverter()
    for work in works:
        total_results += 1
        apc_charges = work.source.apc.charges
        apc_currency = work.source.apc.currency
        if not apc_charges or not apc_currency or apc_currency in ["IRR", "NGN"]:
            continue
        usd_charges = currency_converter.convert(apc_charges, apc_currency, "USD")
        data[work.year_published] += int(usd_charges)
        total_apc += usd_charges
    plot = [{"x": year, "y": value} for year, value in data.items()]
    return {
        "plot": sorted(plot, key=lambda x: -x.get("x")),
        "total_apc": int(total_apc),
        "total_results": total_results,
    }



from typing import Tuple

from flask import Blueprint, request, jsonify, Response
from sentry_sdk import capture_exception

from domain.models.base_model import QueryParams
from domain.services import (
    work_service,
    person_service,
    affiliation_service,
    other_work_service,
    project_service,
    patent_service,
)

search_app_router = Blueprint("search_app_router", __name__)

"""
@api {get} /app/search/person Search persons
@apiName SearchPersons
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de autores por palabra clave.
"""


@search_app_router.route("/person", methods=["GET"])
def search_persons() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = person_service.search_persons(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/works Search works
@apiName SearchWorks
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de productos bibliográficos por palabra clave.
"""


@search_app_router.route("/works", methods=["GET"])
def search_works() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = work_service.search_works(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/works/filters Search works filters
@apiName SearchWorksFilters
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Filtros disponibles en la búsqueda de productos bibliográficos por palabra clave.
"""


@search_app_router.route("/works/filters", methods=["GET"])
def get_search_works_filters() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = work_service.get_search_works_available_filters(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/affiliations/:affiliation_type Search affiliations
@apiName SearchAffiliations
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de afiliaciones por palabra clave y tipo.

@apiParam {String} affiliation_type Tipo de afiliación (ej. "institution", "department").
"""


@search_app_router.route("/affiliations/<affiliation_type>", methods=["GET"])
def search_affiliations(affiliation_type: str) -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = affiliation_service.search_affiliations(affiliation_type, query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/other_works Search other works
@apiName SearchOtherWorks
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de productos de otro tipo por palabra clave.
"""


@search_app_router.route("/other_works", methods=["GET"])
def search_other_works() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = other_work_service.search_other_works(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/patents Search patents
@apiName SearchPatents
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de patentes por palabra clave.
"""


@search_app_router.route("/patents", methods=["GET"])
def search_patents() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = patent_service.search_patents(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400


"""
@api {get} /app/search/projects Search projects
@apiName SearchProjects
@apiGroup Search
@apiVersion 1.0.0

@apiDescription Búsqueda de proyectos por palabra clave.
"""


@search_app_router.route("/projects", methods=["GET"])
def search_projects() -> Response | Tuple[Response, int]:
    try:
        query_params = QueryParams(**request.args)
        data = project_service.search_projects(query_params)
        return jsonify(data)
    except Exception as e:
        capture_exception(e)
        return jsonify({"error": str(e)}), 400



from bson import ObjectId

from domain.models.calculations_model import Calculations
from infrastructure.mongo import calculations_database


def get_person_calculations(person_id: str) -> Calculations:
    person_calculations = calculations_database["person"].find_one({"_id": ObjectId(person_id)})
    if not person_calculations:
        return Calculations()
    return Calculations(**person_calculations)


def get_affiliation_calculations(affiliation_id: str) -> Calculations:
    affiliation_calculations = calculations_database["affiliations"].find_one({"_id": affiliation_id})
    if not affiliation_calculations:
        return Calculations()
    return Calculations(**affiliation_calculations)



import os

import sentry_sdk
from flask import Flask
from flask_compress import Compress
from flask_cors import CORS
from sentry_sdk.integrations.flask import FlaskIntegration

from application.routes.router import router
from config import Settings


def create_app() -> Flask:
    app_settings = Settings()
    sentry_sdk.init(
        dsn=app_settings.SENTRY_DSN,
        integrations=[FlaskIntegration()],
        traces_sample_rate=1.0,
        profiles_sample_rate=1.0,
    )

    project_dir = os.path.dirname(os.path.abspath(__file__))
    static_dir = os.path.join(project_dir, "application", "static")
    app_factory = Flask(__name__, static_folder=static_dir)
    CORS(app_factory)
    app_factory.register_blueprint(router)
    Compress(app_factory)
    return app_factory


if __name__ == "__main__":
    settings = Settings()
    app = create_app()
    app.run(host="0.0.0.0", port=settings.APP_PORT, debug=settings.APP_DEBUG, threaded=True)



from domain.models.other_work_model import OtherWork


def parse_search_results(other_works: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "year_published",
        "title",
        "external_ids",
        "external_urls",
    ]
    return [other_work.model_dump(include=include) for other_work in other_works]


def parse_other_works_by_entity(other_works: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "year_published",
        "title",
        "external_ids",
    ]
    return [other_work.model_dump(include=include) for other_work in other_works]


def parse_other_work(other_work: OtherWork) -> dict:
    return other_work.model_dump()



from domain.models.patent_model import Patent


def parse_search_results(patents: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "title",
        "external_ids",
        "external_urls",
    ]
    return [patent.model_dump(include=include) for patent in patents]


def parse_patents_by_entity(patents: list) -> list:
    include = [
        "id",
        "authors",
        "authors_count",
        "product_types",
        "title",
        "external_ids",
    ]
    return [patent.model_dump(include=include) for patent in patents]


def parse_patent(patent: Patent) -> dict:
    return patent.model_dump()



from typing import Any

from bson import ObjectId
from pydantic import BaseModel, Field

from domain.models.base_model import (
    PyObjectId,
    CitationsCount,
    Type,
    ExternalId,
    ExternalUrl,
    Subject,
    Updated,
    Ranking,
    Publisher,
    APC,
    Author,
    Group,
    Title,
    Affiliation,
    ProductType,
    Name,
)


class BiblioGraphicInfo(BaseModel):
    bibtex: str | None = None
    issue: str | None = None
    pages: str | None = None
    start_page: str | None = None
    end_page: str | None = None
    volume: str | int | None = None


class OpenAccess(BaseModel):
    is_open_access: bool | None = None
    open_access_status: str | None = None
    url: str | None = None
    has_repository_fulltext: bool | None = None


class CitationByYear(BaseModel):
    cited_by_count: int | None
    year: int | None


class Source(BaseModel):
    id: PyObjectId | str | None = None
    name: str | Any | None = None

    issn_l: str | None = None
    is_in_doaj: bool | None = None
    types: list[Type] | None = None
    names: list[Name] | None = None
    scimago_quartile: str | None = None
    serials: dict | None = None
    external_urls: list[ExternalUrl] | None = None
    external_ids: list[ExternalId] | None = None
    ranking: list[Ranking] | None = None
    publisher: Publisher | str | None = None
    apc: APC | None = None
    updated: list[Updated] | None = None

    class Config:
        json_encoders = {ObjectId: str}


class Abstract(BaseModel):
    abstract: dict | None = None
    lang: str | None = None
    provenance: str | None = None
    source: str | None = None


class AutorWork(Author):
    """
    The difference with Author is that in the work the author is marked with a type of author (ex: advisor, co-advisor, author etc..)
    this value is only assigned in the work.
    """

    type: str | None = None


class Work(BaseModel):
    id: PyObjectId = Field(alias="_id")
    abstracts: list[Abstract] | None = None
    apc: APC | None = Field(default_factory=APC)
    authors_count: int | None = Field(default_factory=int, alias="author_count")
    authors: list[AutorWork] | str | None = Field(default_factory=list[AutorWork])
    bibliographic_info: BiblioGraphicInfo | None = None
    citations: list | None = Field(default_factory=list)
    citations_by_year: list[CitationByYear] | None = None
    citations_count: list[CitationsCount] | str | None = None
    date_published: int | None = None
    external_ids: list[ExternalId] | None = None
    external_urls: list[ExternalUrl] | None = None
    groups: list[Group] | str | None = None
    keywords: list | None = None
    open_access: OpenAccess | None = Field(default_factory=OpenAccess)
    ranking: list[Ranking] | str | None = None
    references: list | None = None
    references_count: int | None = None
    source: Source | None = None
    subjects: list[Subject] | str | None = None
    subtitle: str | None = None
    titles: list[Title] | None = None
    types: list[Type] | None = None
    updated: list[Updated] | None = None
    year_published: int | str | None = None
    abstract: dict | None = None
    authors_data: list[Author] | None = None
    affiliations_data: list[Affiliation] | None = None
    countries: str | None = None
    openalex_citations_count: str | int | None = None
    scholar_citations_count: str | int | None = None
    institutions: str | None = None
    departments: str | None = None
    faculties: str | None = None
    groups_ranking: str | None = None
    pages: str | None = None
    start_page: str | None = None
    end_page: str | None = None
    bibtex: str | None = None
    doi: str | None = None
    issue: str | None = None
    language: str | None = None
    open_access_status: str | None = None
    product_types: list[ProductType] | None = None
    publisher: str | None = None
    openalex_types: str | None = None
    scienti_types: str | None = None
    scimago_quartile: str | None = None
    source_apc: str | None = None
    source_data: Source | list[Source] | None = None
    source_name: str | None = None
    source_urls: str | None = None
    title: str | None = None
    volume: str | int | None = None

    class Config:
        json_encoders = {ObjectId: str}



from domain.constants.articles_types import articles_types_list
from domain.models.base_model import QueryParams
from infrastructure.repositories import (
    plot_repository,
    work_repository,
    calculations_repository,
    person_repository,
    affiliation_repository,
)
from domain.parsers import pie_parser, map_parser, venn_parser, bar_parser, network_parser


def get_person_plot(person_id: str, query_params: QueryParams) -> dict:
    return globals()["plot_" + query_params.plot](person_id, query_params)


def plot_annual_evolution_by_scienti_classification(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"project": ["year_published", "types"]}
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return bar_parser.parse_annual_evolution_by_scienti_classification(works)


def plot_annual_citation_count(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"project": ["citations_by_year"]}
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return bar_parser.parse_annual_citation_count(works)


def plot_annual_apc_expenses(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {"source_project": ["apc"], "work_project": ["source", "year_published"]}
    works = work_repository.get_works_with_source_by_person(person_id, query_params, pipeline_params)
    return bar_parser.parse_annual_apc_expenses(works)


def plot_annual_articles_open_access(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "project": ["year_published", "open_access"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return bar_parser.parse_annual_articles_open_access(works)


def plot_annual_articles_by_top_publishers(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["publisher", "apc"],
        "work_project": ["source", "year_published", "types"],
        "match": {
            "types.type": {"$in": articles_types_list},
            "source.publisher.name": {"$ne": float("nan")},
        },
    }
    works = work_repository.get_works_with_source_by_person(person_id, query_params, pipeline_params)
    return bar_parser.parse_annual_articles_by_top_publishers(works)


def plot_most_used_title_words(person_id: str, query_params: QueryParams) -> dict:
    data = calculations_repository.get_person_calculations(person_id)
    return pie_parser.parse_most_used_title_words(data)


def plot_articles_by_publisher(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["publisher"],
        "work_project": ["source"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_publisher(works)


def plot_products_by_subject(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"subjects": {"$ne": []}},
        "project": ["subjects"],
    }
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_products_by_subject(works)


def plot_products_by_database(person_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_products_by_database_by_person(person_id)
    return venn_parser.parse_products_by_database(data)


def plot_articles_by_access_route(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"types.type": {"$in": articles_types_list}},
        "project": ["open_access"],
    }
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_products_by_access_route(works)


def plot_articles_by_scienti_category(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "match": {"types.type": {"$in": articles_types_list}},
        "project": ["ranking"],
    }
    works = work_repository.get_works_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_scienti_category(list(works))


def plot_articles_by_scimago_quartile(person_id: str, query_params: QueryParams) -> dict:
    pipeline_params = {
        "source_project": ["ranking"],
        "work_project": ["source", "date_published"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_scimago_quartile(works)


def plot_articles_by_publishing_institution(person_id: str, query_params: QueryParams) -> dict:
    person = person_repository.get_person_by_id(person_id)
    institution = None
    for affiliation in person.affiliations:
        if any(
            affiliation_type.type not in ["faculty", "department", "group"] for affiliation_type in affiliation.types
        ):
            institution = affiliation_repository.get_affiliation_by_id(str(affiliation.id))
            break
    pipeline_params = {
        "source_project": ["publisher"],
        "work_project": ["source"],
        "match": {"types.type": {"$in": articles_types_list}},
    }
    works = work_repository.get_works_with_source_by_person(person_id, query_params, pipeline_params)
    return pie_parser.parse_articles_by_publishing_institution(works, institution)


def plot_coauthorship_by_country_map(person_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_coauthorship_by_country_map_by_person(person_id, query_params)
    return map_parser.parse_coauthorship_by_country_map(data)


def plot_coauthorship_by_colombian_department_map(person_id: str, query_params: QueryParams) -> dict:
    data = plot_repository.get_coauthorship_by_colombian_department_map_by_person(person_id, query_params)
    return map_parser.get_coauthorship_by_colombian_department_map(data)


def plot_author_coauthorship_network(person_id: str, query_params: QueryParams) -> dict:
    data = calculations_repository.get_person_calculations(person_id)
    return network_parser.parse_institutional_coauthorship_network(data)



from domain.models.person_model import Person


def parse_search_result(persons: list) -> list:
    include = [
        "id",
        "full_name",
        "affiliations",
        "external_ids",
        "products_count",
        "citations_count",
        "logo",
    ]
    return [person.model_dump(include=include) for person in persons]


def parse_person(person: Person) -> dict:
    include = [
        "id",
        "full_name",
        "affiliations",
        "external_ids",
        "products_count",
        "citations_count",
        "logo",
    ]
    return person.model_dump(include=set(include), exclude_none=True)



def parse_search_result(affiliations: list) -> list:
    include = [
        "id",
        "addresses",
        "affiliations",
        "external_ids",
        "external_urls",
        "products_count",
        "citations_count",
        "logo",
        "name",
        "types",
        "products_count",
    ]
    return [affiliation.model_dump(include=include, exclude_none=True) for affiliation in affiliations]
